from enum import Enum
from typing import Annotated
from typing import Optional

from pydantic import BaseModel
from pydantic import Field
from pydantic.types import DirectoryPath


class HridHandling(str, Enum):
    """Enum determining how the HRID generation should be handled.
        - default: Enumerates the HRID, building on the current value in the HRID settings
        - preserve001: Takes the 001 and uses this as the HRID.

    Args:
        str (_type_): _description_
        Enum (_type_): _description_
    """

    default = "default"
    preserve001 = "preserve001"


class FileDefinition(BaseModel):
    file_name: Annotated[
        str,
        Field(
            title="File name",
            description=(
                "Name of the file to be processed. "
                "The location of the file depends on the context"
            ),
        ),
    ] = ""
    discovery_suppressed: Annotated[bool, Field(title="Discovery suppressed")] = False
    staff_suppressed: Annotated[bool, Field(title="Staff suppressed")] = False
    service_point_id: Annotated[str, Field(title="Service point ID")] = ""


class IlsFlavour(str, Enum):
    """ """

    aleph = "aleph"
    voyager = "voyager"
    sierra = "sierra"
    millennium = "millennium"
    koha = "koha"
    tag907y = "tag907y"
    tag001 = "tag001"
    tagf990a = "tagf990a"
    custom = "custom"
    none = "none"


class FolioRelease(str, Enum):
    lotus = "lotus"
    morning_glory = "morning-glory"
    nolana = "nolana"
    orchid = "orchid"
    poppy = "poppy"
    quesnelia = "quesnelia"
    ramsons = "ramsons"
    sunflower = "sunflower"


class LibraryConfiguration(BaseModel):
    okapi_url: str
    tenant_id: str
    okapi_username: str
    okapi_password: str
    base_folder: DirectoryPath = Field(
        description=(
            "The base folder for migration. "
            "Should ideally be a github clone of the migration_repo_template"
        )
    )
    multi_field_delimiter: Optional[str] = "<delimiter>"
    failed_records_threshold: Annotated[
        int, Field(description=("Number of failed records until the process shuts down"))
    ] = 5000
    failed_percentage_threshold: Annotated[
        int, Field(description=("Percentage of failed records until the process shuts down"))
    ] = 20
    library_name: str
    log_level_debug: bool
    folio_release: FolioRelease = Field(
        description=(
            "The Flavour of the ILS you are migrating from. This choice is "
            "maninly tied to the handling of legacy identifiers and thereby the "
            "deterministic UUIDs generated from them."
        )
    )
    iteration_identifier: str
    add_time_stamp_to_file_names: Optional[bool] = False



import csv
import itertools
import json
import logging
import re
import uuid
from functools import reduce
from pathlib import Path
from typing import Dict
from typing import List
from typing import Set
from uuid import UUID

import i18n
from folio_uuid.folio_uuid import FOLIONamespaces
from folio_uuid.folio_uuid import FolioUUID
from folioclient import FolioClient

from folio_migration_tools.custom_exceptions import TransformationFieldMappingError
from folio_migration_tools.custom_exceptions import TransformationProcessError
from folio_migration_tools.custom_exceptions import TransformationRecordFailedError
from folio_migration_tools.library_configuration import LibraryConfiguration
from folio_migration_tools.mapper_base import MapperBase
from folio_migration_tools.mapping_file_transformation.ref_data_mapping import (
    RefDataMapping,
)
from folio_migration_tools.migration_report import MigrationReport

empty_vals = ["Not mapped", None, ""]


class MappingFileMapperBase(MapperBase):
    def __init__(
        self,
        folio_client: FolioClient,
        schema,
        record_map,
        statistical_codes_map,
        uuid_namespace: UUID,
        library_configuration: LibraryConfiguration,
        ignore_legacy_identifier=False,
    ):
        super().__init__(library_configuration, folio_client)
        self.uuid_namespace = uuid_namespace
        self.ignore_legacy_identifier = ignore_legacy_identifier
        self.schema = schema
        self.unique_record_ids: Set[str] = set()

        self.total_records = 0
        self.record_map = record_map
        self.ref_data_dicts: Dict = {}
        self.empty_vals = empty_vals
        self.folio_keys = self.get_mapped_folio_properties_from_map(self.record_map)
        self.field_map = self.setup_field_map(ignore_legacy_identifier)
        self.validate_map()
        try:
            self.mapped_from_values = {
                k["folio_field"]: k["value"]
                for k in self.record_map["data"]
                if k["value"] not in [None, ""] and k["folio_field"] != "legacyIdentifier"
            }
        except KeyError as ke:
            raise TransformationProcessError(
                "",
                "Property missing from one of the settings in the record mapping file",
                f"Property name: {ke}",
            ) from ke

        logging.info(
            "Mapped values:\n%s",
            json.dumps(self.mapped_from_values, indent=4, sort_keys=True),
        )
        legacy_fields = set()
        self.setup_statistical_codes_map(statistical_codes_map)
        self.legacy_record_mappings: dict = {}
        self.mapped_from_legacy_data: dict = {}
        for k in self.record_map["data"]:
            if (
                k["legacy_field"] not in self.empty_vals
                # or k["folio_field"] != "legacyIdentifier"
                or k["value"] not in self.empty_vals
            ):
                clean_folio_field = re.sub(r"\[\d+\]", "", k["folio_field"])
                self.legacy_record_mappings[k["folio_field"]] = list(
                    self.get_map_entries_by_folio_prop_name(
                        clean_folio_field, self.record_map["data"]
                    )
                )
                legacy_fields.add(k["legacy_field"])
                if not self.mapped_from_legacy_data.get(k["folio_field"]):
                    self.mapped_from_legacy_data[k["folio_field"]] = [k["legacy_field"]]
                elif k["legacy_field"] not in self.mapped_from_legacy_data[k["folio_field"]]:
                    self.mapped_from_legacy_data[k["folio_field"]].append(k["legacy_field"])

        logging.info(
            "Mapped legacy fields:\n%s",
            json.dumps(list(legacy_fields), indent=4, sort_keys=True),
        )
        logging.info(
            "Mapped FOLIO fields:\n%s",
            json.dumps(self.folio_keys, indent=4, sort_keys=True),
        )
        csv.register_dialect("tsv", delimiter="\t")

    def setup_statistical_codes_map(self, statistical_codes_map):
        if statistical_codes_map:
            self.statistical_codes_mapping = RefDataMapping(
                self.folio_client,
                "/statistical-codes",
                "statisticalCodes",
                statistical_codes_map,
                "code",
                "StatisticalCodeMapping",
            )
            logging.info("Statistical codes mapping set up")
        else:
            self.statistical_codes_mapping = None
            logging.info("Statistical codes map is not set up")

    def setup_field_map(self, ignore_legacy_identifier):
        field_map = {}  # Map of folio_fields and source fields as an array
        for k in self.record_map["data"]:
            if "folio_field" not in k:
                raise TransformationProcessError(
                    "", "Missing folio_field key in mapping", json.dumps(k)
                )
            if "legacy_field" not in k:
                raise TransformationProcessError(
                    "", "Missing legacy_field key in mapping", json.dumps(k)
                )

            if not field_map.get(k["folio_field"]):
                field_map[k["folio_field"]] = [k["legacy_field"]]
            else:
                field_map[k["folio_field"]].append(k["legacy_field"])
        if not ignore_legacy_identifier and "legacyIdentifier" not in field_map:
            raise TransformationProcessError(
                "",
                "property legacyIdentifier is not in map. Add this property "
                "to the mapping file as if it was a FOLIO property",
            )
        if not ignore_legacy_identifier:
            try:
                self.legacy_id_property_names = field_map["legacyIdentifier"]
                logging.info(
                    "Legacy identifier will be mapped from %s",
                    ",".join(self.legacy_id_property_names),
                )
            except Exception as exception:
                raise TransformationProcessError(
                    "",
                    f"property legacyIdentifier not setup in map: "
                    f"{field_map.get('legacyIdentifier', '') ({exception})}",
                ) from exception
            del field_map["legacyIdentifier"]
        return field_map

    def validate_map(self):
        # TODO: Add functionality here to validate that the map is complete.
        # That it maps the required fields etc
        return True

    @staticmethod
    def get_mapped_folio_properties_from_map(the_map):
        return [
            k["folio_field"]
            for k in the_map["data"]
            if (
                k["legacy_field"] not in empty_vals
                # and k["folio_field"] != "legacyIdentifier"
                or k.get("value", "") not in empty_vals
                or isinstance(k.get("value", ""), bool)
            )
        ]

    @staticmethod
    def get_mapped_legacy_properties_from_map(the_map):
        return [
            k["legacy_field"].strip()
            for k in the_map["data"]
            if (k["legacy_field"].strip() not in empty_vals)
        ]

    def instantiate_record(
        self,
        legacy_object: dict,
        index_or_id,
        object_type: FOLIONamespaces,
        accept_duplicate_ids: bool = False,
    ):
        if self.ignore_legacy_identifier:
            return (
                {
                    "id": str(uuid.uuid4()),
                    "type": "object",
                },
                index_or_id,
            )

        if not (
            legacy_id := " ".join(
                legacy_object.get(li, "") for li in self.legacy_id_property_names
            ).strip()
        ):
            raise TransformationRecordFailedError(
                index_or_id,
                "Could not get a value from legacy object from the property "
                f"{self.legacy_id_property_names}. Check mapping and data",
            )
        generated_id = str(
            FolioUUID(
                self.folio_client.okapi_url,
                object_type,
                legacy_id,
            )
        )
        if generated_id in self.unique_record_ids and not accept_duplicate_ids:
            raise TransformationRecordFailedError(
                index_or_id,
                "Legacy id already generated.",
                f"UUID: {generated_id}, seed: {legacy_id}",
            )
        else:
            self.unique_record_ids.add(generated_id)
        return (
            {
                "id": generated_id,
                "type": "object",
            },
            legacy_id,
        )

    def get_statistical_code(self, legacy_item: dict, folio_prop_name: str, index_or_id):
        if self.statistical_codes_mapping:
            return self.get_mapped_ref_data_value(
                self.statistical_codes_mapping,
                legacy_item,
                index_or_id,
                folio_prop_name,
                True,
            )
        self.migration_report.add(
            "StatisticalCodeMapping",
            i18n.t("Mapping not setup"),
        )
        return ""

    def get_prop(self, legacy_object, folio_prop_name, index_or_id, schema_default_value):
        legacy_item_keys = self.mapped_from_legacy_data.get(folio_prop_name, [])
        map_entries = list(
            MappingFileMapperBase.get_map_entries_by_folio_prop_name(
                folio_prop_name, self.record_map["data"]
            )
        )
        if not any(map_entries):
            return ""
        elif len(map_entries) > 1:
            self.migration_report.add(
                "Details", i18n.t("%{props} were concatenated", props=legacy_item_keys)
            )
            return " ".join(
                MappingFileMapperBase.get_legacy_value(
                    legacy_object,
                    map_entry,
                    self.migration_report,
                    index_or_id,
                    self.library_configuration.multi_field_delimiter,
                )
                for map_entry in map_entries
            ).strip()
        else:
            legacy_value = MappingFileMapperBase.get_legacy_value(
                legacy_object,
                map_entries[0],
                self.migration_report,
                index_or_id,
                self.library_configuration.multi_field_delimiter,
            )
            if legacy_value or isinstance(legacy_value, bool):
                return legacy_value
            else:
                self.migration_report.add(
                    "FolioDefaultValuesAdded",
                    i18n.t(
                        "%{schema_value} added to %{prop_name}",
                        schema_value=schema_default_value,
                        prop_name=folio_prop_name,
                    ),
                )
                return schema_default_value

    def do_map(
        self,
        legacy_object,
        index_or_id: str,
        object_type: FOLIONamespaces,
        accept_duplicate_ids=False,
    ) -> tuple[dict, str]:
        folio_object, legacy_id = self.instantiate_record(
            legacy_object, index_or_id, object_type, accept_duplicate_ids
        )
        for property_name, property in self.schema["properties"].items():
            try:
                self.map_property(property_name, property, folio_object, legacy_id, legacy_object)
            except TransformationFieldMappingError as data_error:
                self.handle_transformation_field_mapping_error(legacy_id, data_error)
        clean_folio_object = self.validate_required_properties(
            legacy_id, folio_object, self.schema, object_type
        )
        return (clean_folio_object, legacy_id)

    def map_property(
        self, schema_property_name: str, schema_property, folio_object, index_or_id, legacy_object
    ):
        if skip_property(schema_property_name, schema_property):
            pass
        elif schema_property.get("type", "") == "object":
            if "properties" in schema_property:
                self.map_object_props(
                    legacy_object,
                    schema_property_name,
                    schema_property,
                    folio_object,
                    index_or_id,
                    1,
                )
        elif schema_property.get("type", "") == "array":
            try:
                if schema_property["items"].get("type", "") == "object":
                    self.map_objects_array_props(
                        legacy_object,
                        schema_property_name,
                        schema_property["items"]["properties"],
                        folio_object,
                        index_or_id,
                        schema_property["items"].get("required", []),
                    )
                    self.validate_object_items_in_array(
                        folio_object, schema_property_name, schema_property
                    )

                elif schema_property["items"].get("type", "") in ["string", "number", "integer"]:
                    self.map_string_array_props(
                        legacy_object,
                        schema_property_name,
                        folio_object,
                        index_or_id,
                    )
                else:
                    logging.info("Edge case %s", schema_property_name)

            except KeyError as schema_anomaly:
                logging.error(
                    "Cannot create property '%s'. Unsupported schema format: %s",
                    schema_property_name,
                    schema_anomaly,
                )

        else:  # Basic property
            self.map_basic_props(
                legacy_object, schema_property_name, folio_object, index_or_id, schema_property
            )

    @staticmethod
    def get_legacy_value(
        legacy_object: dict,
        mapping_file_entry: dict,
        migration_report: MigrationReport,
        index_or_id: str = "",
        multi_field_delimiter="",
    ):
        # Mapping from value fields has preceedence and does not get involved in post processing
        if mapping_file_entry.get("value", "") or isinstance(
            mapping_file_entry.get("value", ""), bool
        ):
            value_mapped_value = mapping_file_entry.get("value")
            migration_report.add(
                "DefaultValuesAdded",
                i18n.t(
                    "%{value} added to %{entry}",
                    value=value_mapped_value,
                    entry=mapping_file_entry.get("folio_field", ""),
                ),
            )
            return value_mapped_value

        # Value mapped from the Legacy field(s)
        value = legacy_object.get(mapping_file_entry["legacy_field"], "")

        if value and mapping_file_entry.get("rules", {}).get("replaceValues", {}):
            if multi_field_delimiter and multi_field_delimiter in value:
                replaced_split_values = [
                    mapping_file_entry["rules"]["replaceValues"].get(sv, "")
                    for sv in value.split(multi_field_delimiter)
                ]
                replaced_val = multi_field_delimiter.join(replaced_split_values)
            else:
                replaced_val = mapping_file_entry["rules"]["replaceValues"].get(value, "")

            if replaced_val or isinstance(replaced_val, bool):
                migration_report.add(
                    "FieldMappingDetails",
                    (
                        f"Replaced {value} in {mapping_file_entry['legacy_field']} "
                        f"with {replaced_val}"
                    ),
                )
                value = replaced_val
        if value and mapping_file_entry.get("rules", {}).get("regexGetFirstMatchOrEmpty", ""):
            my_pattern = (
                f'{mapping_file_entry.get("rules", {}).get("regexGetFirstMatchOrEmpty")}|$'
            )
            value = re.findall(my_pattern, value)[0]
        if not value and mapping_file_entry.get("fallback_legacy_field", ""):
            migration_report.add(
                "FieldMappingDetails",
                (
                    f"Added fallback value from {mapping_file_entry['fallback_legacy_field']} "
                    f"instead of {mapping_file_entry['legacy_field']}"
                ),
            )
            value = legacy_object.get(
                mapping_file_entry.get("fallback_legacy_field", ""), ""
            ).strip()
        if not value and mapping_file_entry.get("fallback_value", ""):
            migration_report.add(
                "FieldMappingDetails",
                (
                    f"Added fallback value {mapping_file_entry['fallback_value']} "
                    f"instead of empty {mapping_file_entry['legacy_field']}"
                ),
            )
            value = mapping_file_entry.get("fallback_value", "")
        return value

    @staticmethod
    def get_legacy_vals(legacy_item, legacy_item_keys):
        result_list = []
        for legacy_item_key in legacy_item_keys:
            val = legacy_item.get(legacy_item_key, "")
            if val not in ["", None]:
                result_list.append(val)
        return result_list

    def map_object_props(
        self,
        legacy_object,
        schema_property_name: str,
        schema_property,
        folio_object,
        index_or_id,
        level: int,
    ):
        temp_object: dict = {}
        for child_property_name, child_property in schema_property["properties"].items():
            sub_prop_path = f"{schema_property_name}.{child_property_name}"
            if "properties" in child_property:
                self.map_object_props(
                    legacy_object,
                    sub_prop_path,
                    child_property,
                    folio_object,
                    index_or_id,
                    level + 1,
                )
            elif (
                child_property.get("type", "") == "array"
                and child_property.get("items", {}).get("type", "") == "object"
                and child_property.get("items", {}).get("properties", "")
            ):
                self.map_objects_array_props(
                    legacy_object,
                    f"{schema_property_name}.{child_property_name}",
                    child_property["items"]["properties"],
                    folio_object,
                    index_or_id,
                    [],
                )
                self.validate_object_items_in_array(
                    legacy_object,
                    child_property_name,
                    child_property["items"]["properties"],
                )

            elif child_property.get("type", "") == "array" and child_property.get("items", {}).get(
                "type", ""
            ) in ["string", "number", "integer"]:
                self.map_string_array_props(
                    legacy_object,
                    f"{schema_property_name}.{child_property_name}",
                    folio_object,
                    index_or_id,
                )
            elif child_property.get("type", "") in ["string", "number", "integer"]:
                path = sub_prop_path.split("].")[-1]
                if p := self.get_prop(
                    legacy_object, sub_prop_path, index_or_id, child_property.get("default", "")
                ):
                    set_deep(folio_object, f"{path}", p)
                # temp_object[child_property_name] = p
            elif p := self.get_prop(
                legacy_object, sub_prop_path, index_or_id, child_property.get("default", "")
            ):
                set_deep(folio_object, sub_prop_path, p)
        if temp_object:
            set_deep(folio_object, schema_property_name, temp_object)
            # folio_object[schema_property_name] = temp_object

    def map_objects_array_props(
        self,
        legacy_object,
        prop_name: str,
        sub_properties,
        folio_object: dict,
        index_or_id,
        required: list[str],
    ):
        resulting_array = []
        i = 0
        while True:
            keys_to_map = {
                k.rsplit(".", 1)[0] for k in self.folio_keys if k.startswith(f"{prop_name}[{i}")
            }
            if not any(keys_to_map):
                break
            for _ in keys_to_map:
                temp_object = {}
                multi_field_props: List[str] = []
                for sub_prop_name, sub_prop in (
                    (k, p)
                    for k, p in sub_properties.items()
                    if not p.get("folio:isVirtual", False)
                ):
                    prop_path = f"{prop_name}[{i}].{sub_prop_name}"
                    if prop_path in self.folio_keys:
                        # We have reached the end of the prop path?
                        res = self.get_prop(
                            legacy_object,
                            prop_path,
                            index_or_id,
                            sub_properties[sub_prop_name].get("default", ""),
                        )
                        self.report_legacy_mapping(
                            self.legacy_basic_property(prop_path), True, True
                        )

                        if (
                            isinstance(res, str)
                            and self.library_configuration.multi_field_delimiter in res
                        ):
                            multi_field_props.append(sub_prop_name)

                        self.validate_enums(res, sub_prop, sub_prop_name, index_or_id, required)
                        if res or isinstance(res, bool):
                            temp_object[sub_prop_name] = res

                    elif (
                        sub_prop_name in sub_properties
                        and sub_properties[sub_prop_name].get("type", "") == "array"
                        and sub_properties[sub_prop_name]["items"].get("type", "") == "object"
                    ):
                        self.map_objects_array_props(
                            legacy_object,
                            prop_path,
                            sub_properties[sub_prop_name]["items"]["properties"],
                            folio_object,
                            index_or_id,
                            [],
                        )
                    elif (
                        sub_prop_name in sub_properties
                        and sub_properties[sub_prop_name].get("type", "") == "array"
                        and sub_properties[sub_prop_name]["items"].get("type", "")
                        in ["string", "number", "integer"]
                    ):
                        # We have not reached the end of the prop path
                        for array_path in [p for p in self.folio_keys if p.startswith(prop_path)]:
                            res = self.get_prop(
                                legacy_object,
                                array_path,
                                index_or_id,
                                sub_properties[sub_prop_name].get("default", ""),
                            )
                            self.validate_enums(
                                res, sub_prop, sub_prop_name, index_or_id, required
                            )
                            if res or isinstance(res, bool):
                                self.add_values_to_string_array(
                                    sub_prop_name,
                                    temp_object,
                                    res,
                                    self.library_configuration.multi_field_delimiter,
                                )

                    elif sub_prop.get("type", "") == "object" and "properties" in sub_prop:
                        self.map_object_props(
                            legacy_object, prop_path, sub_prop, temp_object, index_or_id, 0
                        )
            i = i + 1

            if any(multi_field_props):
                resulting_array.extend(
                    self.split_obj_by_delim(
                        self.library_configuration.multi_field_delimiter,
                        temp_object,
                        multi_field_props,
                    )
                )
            else:
                resulting_array.append(temp_object)

        if any(resulting_array):
            set_deep2(folio_object, prop_name, resulting_array)

    @staticmethod
    def split_obj_by_delim(delimiter: str, folio_obj: dict, delimited_props: List[str]):
        non_split_props = [(k, v) for k, v in folio_obj.items() if k not in delimited_props]
        delimited_props = map(lambda x: [x, *folio_obj[x].split(delimiter)], delimited_props)
        zipped = list(zip(*delimited_props))
        res = []
        for (prop_name_idx, prop_name), (value_idx, ra) in itertools.product(
            enumerate(zipped[0]), enumerate(zipped[1:])
        ):
            if prop_name_idx == 0:
                res.append({prop_name: ra[prop_name_idx]})
            else:
                res[value_idx][prop_name] = ra[prop_name_idx]
        for r in res:
            r.update(non_split_props)
        return res

    def map_string_array_props(self, legacy_object, prop, folio_object, index_or_id):
        keys_to_map = [k for k in self.folio_keys if k.startswith(prop)]
        for prop_name in keys_to_map:
            if prop_name in self.folio_keys and self.has_property(legacy_object, prop_name):
                if mapped_prop := self.get_prop(legacy_object, prop_name, index_or_id, ""):
                    self.add_values_to_string_array(
                        prop,
                        folio_object,
                        mapped_prop,
                        self.library_configuration.multi_field_delimiter,
                    )
                self.report_legacy_mapping(self.legacy_basic_property(prop_name), True, True)

    @staticmethod
    def add_values_to_string_array(prop, folio_object, mapped_prop_value, delimiter: str):
        if in_deep(folio_object, prop) and mapped_prop_value not in get_deep(
            folio_object, prop, []
        ):
            if isinstance(mapped_prop_value, str) and delimiter in mapped_prop_value:
                old_prop = get_deep(folio_object, prop)
                set_deep(folio_object, prop, old_prop.extend(mapped_prop_value.split(delimiter)))

            else:
                old_prop = get_deep(folio_object, prop)
                added_prop = old_prop.append(mapped_prop_value)
                set_deep(folio_object, prop, [added_prop])
        elif isinstance(mapped_prop_value, str) and delimiter in mapped_prop_value:
            set_deep(folio_object, prop, mapped_prop_value.split(delimiter))
        else:
            # No values in array previously
            set_deep(folio_object, prop, [mapped_prop_value])

    def map_basic_props(
        self, legacy_object, property_name, folio_object, index_or_id, schema_property
    ):
        if self.has_basic_property(legacy_object, property_name):  # is there a match in the csv?
            mapped_prop = self.get_prop(
                legacy_object, property_name, index_or_id, schema_property.get("default", "")
            )
            if mapped_prop or isinstance(mapped_prop, bool):
                self.validate_enums(
                    mapped_prop,
                    schema_property,
                    property_name,
                    index_or_id,
                    self.schema.get("required", []),
                )
                folio_object[property_name] = mapped_prop
            self.report_legacy_mapping(self.legacy_basic_property(property_name), True, True)

    @staticmethod
    def _get_delimited_file_reader(source_file, file_name: Path):
        """
            First, let's count:
            * The total number of rows in the source file
            * The total number of empty rows in the source file

            Then, we'll return those counts and a csv.DictReader

        Args:
            source_file (_type_): _description_
            file_name (Path): _description_

        Returns:
            (int, int, DictReader): total rows, empty rows, dict reader
        """
        empty_rows = 0
        total_rows = -1  # Do not count header row
        if str(file_name).endswith("tsv"):
            delimiter = "\t"
        else:
            delimiter = ","
        for line in source_file:
            if not "".join(line.strip().split(delimiter)):  # check for empty rows
                empty_rows += 1
            total_rows += 1
        source_file.seek(0)  # Set file position back to start
        if str(file_name).endswith("tsv"):
            dict_reader = csv.DictReader(source_file, dialect="tsv")
        else:
            dict_reader = csv.DictReader(source_file)
        return total_rows, empty_rows, dict_reader

    def get_objects(self, source_file, file_name: Path):
        total_rows, empty_rows, reader = self._get_delimited_file_reader(source_file, file_name)
        logging.info("Source data file contains %d rows", total_rows)
        logging.info("Source data file contains %d empty rows", empty_rows)
        self.migration_report.set(
            "GeneralStatistics", "Number of rows in {}".format(file_name.name), total_rows
        )
        self.migration_report.set(
            "GeneralStatistics",
            "Number of empty rows in {}".format(file_name.name),
            empty_rows,
        )
        try:
            yield from reader
        except Exception as exception:
            logging.error("%s at row %s", exception, reader.line_num)
            raise exception from exception

    def has_property(self, legacy_object, folio_prop_name: str):
        legacy_keys = self.field_map.get(folio_prop_name, [])
        return (
            any(legacy_keys)
            and any(k not in empty_vals for k in legacy_keys)
            and any(legacy_object.get(legacy_key, "") for legacy_key in legacy_keys)
        )

    def has_basic_property(self, legacy_object, folio_prop_name):
        if folio_prop_name not in self.folio_keys:
            return False
        if folio_prop_name in self.mapped_from_values:
            return True
        legacy_mappings = self.legacy_record_mappings.get(folio_prop_name, [])

        return any(legacy_mappings) and any(
            legacy_mapping not in empty_vals for legacy_mapping in legacy_mappings
        )

    @staticmethod
    def get_map_entries_by_folio_prop_name(folio_prop_name, data):
        return (
            k
            for k in data
            if k["folio_field"] == folio_prop_name
            and any(
                [
                    is_set_or_bool_or_numeric(k.get("value", "")),
                    is_set_or_bool_or_numeric(k.get("legacy_field", "")),
                    is_set_or_bool_or_numeric(k.get("fallback_legacy_field", "")),
                    is_set_or_bool_or_numeric(k.get("fallback_value", "")),
                ]
            )
        )

    def legacy_basic_property(self, folio_prop):
        if folio_prop not in self.folio_keys:
            return ""
        return next(
            (k["legacy_field"] for k in self.record_map["data"] if k["folio_field"] == folio_prop),
            "",
        )

    def verify_legacy_record(self, legacy_object, idx):
        if idx == 0:
            missing_keys_in_record = [
                f
                for f in self.get_mapped_legacy_properties_from_map(self.record_map)
                if f not in legacy_object
            ]
            if any(missing_keys_in_record):
                raise TransformationProcessError(
                    "",
                    ("There are mapped legacy fields that are not in the legacy record"),
                    missing_keys_in_record,
                )
            else:
                logging.info("All maped legacy fields are in the legacy object")

    def get_ref_data_tuple_by_code(self, ref_data, ref_name, code):
        return self.get_ref_data_tuple(ref_data, ref_name, code, "code")

    def get_ref_data_tuple_by_name(self, ref_data, ref_name, name):
        return self.get_ref_data_tuple(ref_data, ref_name, name, "name")

    def get_ref_data_tuple(self, ref_data, ref_name, key_value, key_type):
        dict_key = f"{ref_name}{key_type}"
        if ref_object := self.ref_data_dicts.get(dict_key, {}).get(key_value.lower().strip(), ()):
            return ref_object
        d = {r[key_type].lower(): (r["id"], r["name"]) for r in ref_data}
        self.ref_data_dicts[dict_key] = d
        return self.ref_data_dicts.get(dict_key, {}).get(key_value.lower().strip(), ())

    def validate_enums(
        self,
        mapped_value,
        mapped_schema_property,
        mapped_schema_property_name,
        index_or_id,
        required,
    ):
        if (
            (
                "enum" in mapped_schema_property
                and mapped_value
                and mapped_value not in mapped_schema_property["enum"]
            )
            or (
                "enum" in mapped_schema_property
                and mapped_schema_property_name in required
                and not mapped_value
            )
            or (
                mapped_schema_property.get("items", {}).get("enum")
                and mapped_value
                and mapped_value not in mapped_schema_property.get("items", {}).get("enum")
            )
        ):
            raise TransformationRecordFailedError(
                index_or_id,
                f"Allowed values for {mapped_schema_property_name}"
                f"are {mapped_schema_property['enum']} "
                f"Forbidden enum value found: ",
                mapped_value,
            )

    def is_uuid(self, value):
        """Returns True if the value is a UUID, and False if it is not.

        Args:
            value (_type_): a value that may or may not be a UUID

        Returns:
            bool: True/False
        """
        try:
            uuid.UUID(str(value))
        except ValueError:
            return False
        return True

    def validate_object_items_in_array(self, folio_object, schema_property_name, schema_property):
        valid_array_objects = []
        for item in folio_object.get(schema_property_name, []):
            if all(
                item.get(r) or (isinstance(item.get(r), bool))
                for r in schema_property["items"].get("required", [])
            ):
                valid_array_objects.append(item)
            else:
                self.migration_report.add(
                    "IncompleteSubPropertyRemoved",
                    f"{schema_property_name}",
                )
        if valid_array_objects:
            folio_object[schema_property_name] = valid_array_objects
        else:
            folio_object.pop(schema_property_name, [])


def skip_property(property_name, property):
    return bool(
        property_name in ["metadata", "id", "type", "lastCheckIn"]
        or property_name.startswith("effective")
        or property.get("folio:isVirtual", False)
        or property.get("description", "") == "Deprecated"
    )


def weird_division(number, divisor):
    return number / divisor if divisor else 0


def set_deep(dictionary, key, value):
    """sets a nested property in a dict given a dot notated address

    Args:
        dictionary (_type_): a python dictionary ({"a":{"b":{"c":"value"}}})
        key (_type_): A string of dot notated address (a.b.c)
        value (_type_): the value to set

    """
    dd = dictionary
    keys = key.split(".")
    latest = keys.pop()
    for k in keys:
        dd = dd.setdefault(k, {})
    dd.setdefault(latest, value)


def set_deep2(dictionary, key, value):
    """sets a nested property in a dict given a dot notated address

    Args:
        dictionary (_type_): a python dictionary ({"a":{"b":{"c":"value"}}})
        key (_type_): A string of dot notated address (a.b.c)
        value (_type_): the value to set

    """
    dd = dictionary
    keys = key.split(".")
    latest = keys.pop()
    name = ""
    number = 0
    for k in keys:
        if k == keys[0] and k.endswith("]"):
            m = re.search(r"\[([\d]+)\]", k)
            number = int(m[1])
            name = k.split("[")[0]
            dd = dd.setdefault(name, [{}])
        else:
            dd = dd.setdefault(k, {})
    if name and keys and keys[0].startswith(name):
        if len(dd) <= number:
            dd.append({})
        dd[number][latest] = value
    elif latest in dd:
        for i in range(len(value)):
            if len(dd[latest]) > i and dd[latest][i] and isinstance(dd[latest][i], dict):
                dd[latest][i].update(value[i])
            else:
                dd[latest].insert(i, value[i])

    else:
        dd[latest] = value


def get_deep(dictionary, keys, default=None):
    """returns a nested property in a dict given a dot notated address

    Args:
        dictionary (_type_): a python dictionary ({"a":{"b":{"c":"value"}}})
        keys (_type_): A string of dot notated address (a.b.c)
        default (_type_): Default value to return

    Returns:
        _type_: the value/property of the dict
    """
    return reduce(
        lambda d, key: d.get(key, default) if isinstance(d, dict) else default,
        keys.split("."),
        dictionary,
    )


def in_deep(dictionary, keys):
    """Checks if a property exists given a dot notated address

    Args:
        dictionary (_type_): a python dictionary ({"a":{"b":{"c":"value"}}})
        keys (_type_): A string of dot notated address (a.b.c)

    Returns:
        _type_: a truthy value or False is there is a property in the dict
    """
    return reduce(
        lambda d, key: d.get(key, False) if isinstance(d, dict) else False,
        keys.split("."),
        dictionary,
    )


def is_set_or_bool_or_numeric(any_value):
    return any(isinstance(any_value, t) for t in [int, bool, float, complex]) or any_value.strip()



from typing import Any


class TransactionResult(object):
    __slots__ = [
        "was_successful",
        "folio_loan",
        "should_be_retried",
        "error_message",
        "migration_report_message",
    ]

    def __init__(
        self,
        was_successful: bool,
        should_be_retried: bool,
        folio_loan: Any,
        error_message: str,
        migration_report_message: str,
    ):
        self.was_successful = was_successful
        self.folio_loan = folio_loan
        self.should_be_retried = should_be_retried
        self.error_message = error_message
        self.migration_report_message = migration_report_message



import logging
from typing import Annotated
from typing import List
import i18n

from folio_uuid.folio_namespaces import FOLIONamespaces
from pydantic import Field

from folio_migration_tools.helper import Helper
from folio_migration_tools.library_configuration import FileDefinition
from folio_migration_tools.library_configuration import IlsFlavour
from folio_migration_tools.library_configuration import LibraryConfiguration
from folio_migration_tools.marc_rules_transformation.marc_file_processor import (
    MarcFileProcessor,
)
from folio_migration_tools.marc_rules_transformation.rules_mapper_authorities import (
    AuthorityMapper,
)
from folio_migration_tools.migration_tasks.migration_task_base import MigrationTaskBase
from folio_migration_tools.task_configuration import AbstractTaskConfiguration


class AuthorityTransformer(MigrationTaskBase):
    class TaskConfiguration(AbstractTaskConfiguration):
        name: Annotated[
            str,
            Field(
                description=(
                    "Name of this migration task. The name is being used to call the specific "
                    "task, and to distinguish tasks of similar types"
                )
            ),
        ]
        migration_task_type: Annotated[
            str,
            Field(
                title="Migration task type",
                description=("The type of migration task you want to perform"),
            ),
        ]
        files: Annotated[
            List[FileDefinition],
            Field(
                title="Source files", description=("List of MARC21 files with authority records")
            ),
        ]
        ils_flavour: Annotated[
            IlsFlavour,
            Field(
                title="ILS flavour", description="The type of ILS you are migrating records from."
            ),
        ]
        tags_to_delete: Annotated[
            List[str],
            Field(
                title="Tags to delete from MARC record",
                description=(
                    "Tags in the incoming MARC authority that the process should remove "
                    "before adding them into FOLIO. These tags will be used in the "
                    "transformation before getting removed."
                ),
            ),
        ] = []
        create_source_records: Annotated[
            bool,
            Field(
                title="Create source records",
                description=(
                    "Controls wheter or not to retain the MARC records in "
                    "Source Record Storage."
                ),
            ),
        ] = True

    @staticmethod
    def get_object_type() -> FOLIONamespaces:
        return FOLIONamespaces.authorities

    def __init__(
        self,
        task_config: TaskConfiguration,
        library_config: LibraryConfiguration,
        use_logging: bool = True,
    ):
        super().__init__(library_config, task_config, use_logging)
        self.processor: MarcFileProcessor
        self.check_source_files(
            self.folder_structure.legacy_records_folder, self.task_configuration.files
        )
        self.mapper: AuthorityMapper = AuthorityMapper(
            self.folio_client, library_config, task_config
        )
        self.auth_ids: set = set()
        logging.info("Init done")

    def do_work(self):
        self.do_work_marc_transformer()

    def wrap_up(self):
        logging.info("Done. Transformer Wrapping up...")
        self.extradata_writer.flush()
        self.processor.wrap_up()
        with open(self.folder_structure.migration_reports_file, "w+") as report_file:
            self.mapper.migration_report.write_migration_report(
                i18n.t("Authority records transformation report"),
                report_file,
                self.start_datetime,
            )
            Helper.print_mapping_report(
                report_file,
                self.mapper.parsed_records,
                self.mapper.mapped_folio_fields,
                self.mapper.mapped_legacy_fields,
            )
        logging.info(
            "Done. Transformation report written to %s",
            self.folder_structure.migration_reports_file.name,
        )
        self.clean_out_empty_logs()



import json
import logging
import uuid
from pathlib import Path
from unittest.mock import MagicMock
from unittest.mock import Mock

from folioclient import FolioClient

from folio_migration_tools.extradata_writer import ExtradataWriter
from folio_migration_tools.mapping_file_transformation.holdings_mapper import (
    HoldingsMapper,
)
from folio_migration_tools.migration_report import MigrationReport


def mocked_holdings_mapper() -> Mock:
    mock_mapper = Mock(spec=HoldingsMapper)
    mock_mapper.migration_report = MigrationReport()
    mock_mapper.extradata_writer = ExtradataWriter(Path(""))

    return mock_mapper


def mocked_folio_client() -> FolioClient:
    try:
        FolioClient.login = MagicMock(name="login", return_value=None)
        FolioClient.okapi_token = "token"  # noqa:S105
        mocked_folio = FolioClient("okapi_url", "tenant_id", "username", "password")
        mocked_folio.folio_get_single_object = folio_get_single_object_mocked
        mocked_folio.folio_get_all = folio_get_all_mocked
        mocked_folio.get_from_github = folio_get_from_github
        mocked_folio.current_user = str(uuid.uuid4())
        return mocked_folio
    except Exception as ee:
        logging.error(ee)
        raise ee


def folio_get_all_mocked(ref_data_path, array_name, query="", limit=10):
    with open("./static/reference_data.json", "r") as super_schema_file:
        super_schema = json.load(super_schema_file)
    if ref_data_path == "/coursereserves/terms":
        yield from [
            {"name": "Fall 2022", "id": "42093be3-d1e7-4bb6-b2b9-18e153d109b2"},
            {"name": "Summer 2022", "id": "415b14a8-c94c-4aa1-a0a8-d397efae343e"},
        ]
    elif ref_data_path == "/coursereserves/departments":
        yield from [
            {
                "id": "7532e5ab-9812-496c-ab77-4fbb6a7e5dbf",
                "name": "Department_t",
                "description": "Art & Art History",
            },
            {
                "id": "af7ae6be-c0b2-444d-b76f-4061098d17cd",
                "name": "Department_FALLBACK",
                "description": "FALLBACK",
            },
        ]
    elif ref_data_path == "/organizations-storage/categories":
        yield from [
            {"id": "c78640d5-a1ec-4721-9a1f-c6f876d4c179", "value": "Returns"},
            {"id": "604c2c9d-ed3a-46cd-bec4-69926c303b22", "value": "Sales"},
            {"id": "c5b175bd-34a0-4a4d-9bd9-8eddae8e67f8", "value": "General"},
            {"id": "97dcb23df-1aba-444e-b88d-804d17c715a5", "value": "Technical Support"},
            {"id": "e193b0d1-4674-4a9e-818b-375f013d963f", "value": "Moral Support"},
        ]

    elif ref_data_path == "/organizations-storage/organization-types":
        yield from [
            {"id": "837d04b6-d81c-4c49-9efd-2f62515999b3", "name": "Consortium"},
            {"id": "fc54327d-fd60-4f6a-ba37-a4375511b91b", "name": "Unspecified"},
        ]
    elif (
        ref_data_path == "/organizations-storage/organizations"
        and query == '?query=(code=="EBSCO")'
    ):
        yield from [{"id": "some id", "code": "some code", "name": "EBSCO Information Services"}]

    elif (
        ref_data_path == "/organizations-storage/organizations"
        and query == '?query=(code=="LisasAwesomeStartup")'
    ):
        yield from []

    elif ref_data_path == "/organizations-storage/organizations":
        yield from [
            {"id": "837d04b6-d81c-4c49-9efd-2f62515999b3", "code": "GOBI"},
            {"id": "fc54327d-fd60-4f6a-ba37-a4375511b91b", "code": "EBSCO"},
        ]

    elif ref_data_path == "/orders/acquisition-methods":
        yield from [
            {"id": "837d04b6-d81c-4c49-9efd-2f62515999b3", "value": "Purchase"},
            {"id": "fc54327d-fd60-4f6a-ba37-a4375511b91b", "value": "Theft"},
            {"id": "fc54327d-fd60-4f6a-ba37-a437551sarfs91b", "value": "Other"},
        ]

    elif ref_data_path == "/groups":
        yield from [
            {
                "group": "FOLIO fallback group name",
                "desc": "Mocked response",
                "id": "27ab99d3-0e17-41f0-a20a-99e05acc0e6f",
            },
            {
                "group": "FOLIO group name",
                "desc": "Mocked response",
                "id": "5fc96cbd-a860-42a7-8d2b-72af30206712",
            },
        ]
    elif ref_data_path == "/departments":
        yield from [
            {
                "id": "12a2ad12-951d-4124-9fb2-58c70f0b7f71",
                "name": "FOLIO user department name",
                "code": "fdp",
            },
            {
                "id": "2f452d21-507d-4b32-a89d-8ea9753cc946",
                "name": "FOLIO fallback user department name",
                "code": "fb",
            },
        ]
    elif ref_data_path == "/owners":
        yield from [
            {
                "owner": "The Best Fee Fine Owner",
                "desc": "She really is!",
                "servicePointOwner": [
                    {"value": "a77b55e7-f9f3-40a1-83e0-241bc606a826", "label": "lisatest"}
                ],
                "id": "5abfff3f-50eb-432a-9a43-21f8f7a70194",
            },
            {
                "owner": "The Other Fee Fine Owner",
                "desc": "heeey",
                "servicePointOwner": [
                    {"value": "1543c345-dcaf-4367-84a8-853d95837a3b", "label": "lisatest2 :) <3 "}
                ],
                "id": "62a0eb54-de96-46ee-b184-5be6c8114a19",
            },
        ]
    elif ref_data_path == "/feefines":
        yield from [
            {
                "automatic": False,
                "feeFineType": "Coffee spill",
                "ownerId": "5abfff3f-50eb-432a-9a43-21f8f7a70194",
                "id": "6e8dc178-f667-45cd-90b5-338c78c3a85c",
            },
            {
                "automatic": False,
                "feeFineType": "Coffee spill",
                "ownerId": "62a0eb54-de96-46ee-b184-5be6c8114a19",
                "id": "031836ec-521a-4493-9f76-0e02c2e7d241",
            },
            {
                "automatic": False,
                "feeFineType": "Replacement library card",
                "ownerId": "5abfff3f-50eb-432a-9a43-21f8f7a70194",
                "id": "8936606d-223b-428e-9a70-4b8105f60cdb",
            },
            {
                "automatic": True,
                "feeFineType": "Replacement processing fee",
                "id": "d20df2fb-45fd-4184-b238-0d25747ffdd9",
            },
        ]

    elif ref_data_path == "/service-points":
        yield from [
            {
                "id": "finance_office_uuid",
                "name": "Finance Office",
                "code": "fo",
            },
            {
                "id": "library_main_desk_uuid",
                "name": "Library Main Desk",
                "code": "lmd",
            },
        ]

    elif ref_data_path == "/users" and query == '?query=(externalSystemId=="Some external id")':
        yield from [{"id": "some id", "barcode": "some barcode", "patronGroup": "some group"}]
    elif ref_data_path == "/users" and query == '?query=(barcode=="u123")':
        yield from [{"id": "user123", "barcode": "u123", "patronGroup": "some group"}]
    elif ref_data_path == "/users" and query == '?query=(barcode=="u456")':
        yield from [{"id": "user456", "barcode": "u456", "patronGroup": "some group"}]

    elif ref_data_path == "/inventory/items" and query == '?query=(barcode=="some barcode")':
        yield from [
            {
                "id": "a FOLIO item uuid",
                "title": "Döda fallen i Avesta.",
                "barcode": "some barcode",
                "callNumber": "QB611 .C44",
                "materialType": {
                    "id": "4eea3f27-8910-46fc-9666-e2b44326c2b8",
                    "name": "sound recording",
                },
                "effectiveLocation": {
                    "id": "2e48e713-17f3-4c13-a9f8-23845bb210a4",
                    "name": "Reading room",
                },
            }
        ]

    elif ref_data_path in super_schema:
        yield from super_schema.get(ref_data_path)
    else:
        yield {}


def folio_get_single_object_mocked(*args, **kwargs):
    with open("./static/reference_data.json", "r") as super_schema_file:
        super_schema = json.load(super_schema_file)
    if args[0] == "/hrid-settings-storage/hrid-settings":
        return {
            "instances": {"prefix": "pref", "startNumber": 1},
            "holdings": {"prefix": "pref", "startNumber": 1},
            "items": {"prefix": "pref", "startNumber": 1},
            "commonRetainLeadingZeroes": True,
        }

    elif (
        args[0] == "/configurations/entries?query=(module==ORG%20and%20configName==localeSettings)"
    ):
        return {
            "configs": [
                {
                    "value": '{"timezone":"America/New_York"}',
                }
            ]
        }

    elif args[0] in super_schema:
        return super_schema.get(args[0])


def folio_get_from_github(owner, repo, file_path):
    return FolioClient.get_latest_from_github(owner, repo, file_path, "")



import csv
import json
import logging
import sys
import time
import i18n
from typing import Optional
from zoneinfo import ZoneInfo

from folio_uuid.folio_namespaces import FOLIONamespaces

from folio_migration_tools.circulation_helper import CirculationHelper
from folio_migration_tools.custom_dict import InsensitiveDictReader
from folio_migration_tools.helper import Helper
from folio_migration_tools.library_configuration import FileDefinition
from folio_migration_tools.library_configuration import LibraryConfiguration
from folio_migration_tools.migration_report import MigrationReport
from folio_migration_tools.migration_tasks.migration_task_base import MigrationTaskBase
from folio_migration_tools.task_configuration import AbstractTaskConfiguration
from folio_migration_tools.transaction_migration.legacy_request import LegacyRequest


class RequestsMigrator(MigrationTaskBase):
    class TaskConfiguration(AbstractTaskConfiguration):
        name: str
        migration_task_type: str
        open_requests_file: FileDefinition
        starting_row: Optional[int] = 1
        item_files: Optional[list[FileDefinition]] = []
        patron_files: Optional[list[FileDefinition]] = []

    @staticmethod
    def get_object_type() -> FOLIONamespaces:
        return FOLIONamespaces.requests

    def __init__(
        self,
        task_configuration: TaskConfiguration,
        library_config: LibraryConfiguration,
    ):
        csv.register_dialect("tsv", delimiter="\t")
        self.migration_report = MigrationReport()
        self.valid_legacy_requests = []
        super().__init__(library_config, task_configuration)
        self.circulation_helper = CirculationHelper(
            self.folio_client,
            "",
            self.migration_report,
        )
        try:
            logging.info("Attempting to retrieve tenant timezone configuration...")
            my_path = (
                "/configurations/entries?query=(module==ORG%20and%20configName==localeSettings)"
            )
            self.tenant_timezone_str = json.loads(
                self.folio_client.folio_get_single_object(my_path)["configs"][0]["value"]
            )["timezone"]
            logging.info("Tenant timezone is: %s", self.tenant_timezone_str)
        except Exception:
            logging.info('Tenant locale settings not available. Using "UTC".')
            self.tenant_timezone_str = "UTC"
        self.tenant_timezone = ZoneInfo(self.tenant_timezone_str)
        with open(
            self.folder_structure.legacy_records_folder
            / task_configuration.open_requests_file.file_name,
            "r",
            encoding="utf-8",
        ) as requests_file:
            self.semi_valid_legacy_requests = list(
                self.load_and_validate_legacy_requests(
                    InsensitiveDictReader(requests_file, dialect="tsv")
                )
            )
            logging.info(
                "Loaded and validated %s requests in file",
                len(self.semi_valid_legacy_requests),
            )
        if any(self.task_configuration.item_files) or any(self.task_configuration.patron_files):
            self.valid_legacy_requests = list(self.check_barcodes())
            logging.info(
                "Loaded and validated %s requests against barcodes",
                len(self.valid_legacy_requests),
            )
        else:
            logging.info(
                "No item or user files supplied. Not validating against"
                "previously migrated objects"
            )
            self.valid_legacy_requests = self.semi_valid_legacy_requests

        self.valid_legacy_requests.sort(key=lambda x: x.request_date)
        logging.info("Sorted the list of requests by request date")

        self.t0 = time.time()
        self.skipped_since_already_added = 0
        self.failed_requests = set()
        logging.info("Starting row is %s", task_configuration.starting_row)
        logging.info("Init completed")

    def prepare_legacy_request(self, legacy_request: LegacyRequest):
        patron = self.circulation_helper.get_user_by_barcode(legacy_request.patron_barcode)
        self.migration_report.add_general_statistics(i18n.t("Patron lookups performed"))

        if not patron:
            logging.error(f"No user with barcode {legacy_request.patron_barcode} found in FOLIO")
            Helper.log_data_issue(
                f"{legacy_request.patron_barcode}",
                "No user with barcode.",
                f"{legacy_request.patron_barcode}",
            )
            self.migration_report.add_general_statistics(
                i18n.t("No user with barcode found in FOLIO")
            )
            self.failed_requests.add(legacy_request)
            return False, legacy_request
        legacy_request.patron_id = patron.get("id")

        item = self.circulation_helper.get_item_by_barcode(legacy_request.item_barcode)
        self.migration_report.add_general_statistics(i18n.t("Item lookups performed"))
        if not item:
            logging.error(f"No item with barcode {legacy_request.item_barcode} found in FOLIO")
            self.migration_report.add_general_statistics(
                i18n.t("No item with barcode found in FOLIO")
            )
            Helper.log_data_issue(
                f"{legacy_request.item_barcode}",
                "No item with barcode",
                f"{legacy_request.item_barcode}",
            )
            self.failed_requests.add(legacy_request)
            return False, legacy_request
        holding = self.circulation_helper.get_holding_by_uuid(item.get("holdingsRecordId"))
        self.migration_report.add_general_statistics(i18n.t("Holdings lookups performed"))
        legacy_request.item_id = item.get("id")
        legacy_request.holdings_record_id = item.get("holdingsRecordId")
        legacy_request.instance_id = holding.get("instanceId")
        if item["status"]["name"] in ["Available"]:
            legacy_request.request_type = "Page"
            logging.info(f'Setting request to Page, since the status is {item["status"]["name"]}')
        self.migration_report.add_general_statistics(
            i18n.t("Valid, prepared requests, ready for posting")
        )
        return True, legacy_request

    def do_work(self):
        logging.info("Starting")
        if self.task_configuration.starting_row > 1:
            logging.info(f"Skipping {(self.task_configuration.starting_row-1)} records")
        for num_requests, legacy_request in enumerate(
            self.valid_legacy_requests[self.task_configuration.starting_row - 1 :],
            start=1,
        ):
            t0_migration = time.time()
            try:
                res, legacy_request = self.prepare_legacy_request(legacy_request)
                if res:
                    if self.circulation_helper.create_request(
                        self.folio_client, legacy_request, self.migration_report
                    ):
                        self.migration_report.add_general_statistics(
                            i18n.t("Successfully migrated requests")
                        )
                    else:
                        self.migration_report.add_general_statistics(
                            i18n.t("Unsuccessfully migrated requests")
                        )
                        self.failed_requests.add(legacy_request)
                if num_requests == 1:
                    logging.info(json.dumps(legacy_request.to_dict(), indent=4))
            except Exception:
                logging.exception(
                    "Error in row %s  Item barcode: %s Patron barcode: %s",
                    num_requests,
                    legacy_request.item_barcode,
                    legacy_request.patron_barcode,
                )
                sys.exit(1)
            if num_requests % 10 == 0:
                logging.info(f"{timings(self.t0, t0_migration, num_requests)} {num_requests}")
        logging.info(f"{timings(self.t0, t0_migration, num_requests)} {num_requests}")

    def wrap_up(self):
        self.extradata_writer.flush()
        self.write_failed_request_to_file()

        with open(self.folder_structure.migration_reports_file, "w+") as report_file:
            self.migration_report.write_migration_report(
                i18n.t("Requests migration report"), report_file, self.start_datetime
            )
        self.clean_out_empty_logs()

    def write_failed_request_to_file(self):
        csv_columns = [
            "item_barcode",
            "patron_barcode",
            "request_date",
            "request_expiration_date",
            "comment",
            "request_type",
            "pickup_servicepoint_id",
        ]
        with open(self.folder_structure.failed_recs_path, "w+") as failed_requests_file:
            writer = csv.DictWriter(failed_requests_file, fieldnames=csv_columns, dialect="tsv")
            writer.writeheader()
            failed: LegacyRequest
            for failed in self.failed_requests:
                writer.writerow(failed.to_source_dict())

    def check_barcodes(self):
        user_barcodes = set()
        item_barcodes = set()
        self.circulation_helper.load_migrated_item_barcodes(
            item_barcodes, self.task_configuration.item_files, self.folder_structure
        )
        self.circulation_helper.load_migrated_user_barcodes(
            user_barcodes, self.task_configuration.patron_files, self.folder_structure
        )

        request: LegacyRequest
        for request in self.semi_valid_legacy_requests:
            has_item_barcode = request.item_barcode in item_barcodes
            has_patron_barcode = request.patron_barcode in user_barcodes
            if has_item_barcode and has_patron_barcode:
                self.migration_report.add_general_statistics(
                    i18n.t("Requests successfully verified against migrated users and items")
                )
                yield request
            else:
                self.migration_report.add(
                    "DiscardedLoans",
                    i18n.t(
                        "Requests discarded. Had migrated item barcode: %{item_barcode}.\n Had migrated user barcode: %{patron_barcode}",
                        item_barcode=has_item_barcode,
                        patron_barcode=has_patron_barcode,
                    ),
                )
                self.migration_report.add_general_statistics(
                    i18n.t("Requests that failed verification against migrated users and items")
                )
            if not has_item_barcode:
                Helper.log_data_issue(
                    "",
                    "Request without matched item barcode",
                    json.dumps(request.to_source_dict()),
                )
            if not has_patron_barcode:
                Helper.log_data_issue(
                    "",
                    "Request without matched patron barcode",
                    json.dumps(request.to_source_dict()),
                )

    def load_and_validate_legacy_requests(self, requests_reader):
        num_bad = 0
        logging.info("Validating legacy requests in file...")
        for legacy_reques_count, legacy_request_dict in enumerate(requests_reader, start=1):
            self.migration_report.add_general_statistics(i18n.t("Requests in file"))
            try:
                legacy_request = LegacyRequest(
                    legacy_request_dict,
                    self.tenant_timezone,
                    legacy_reques_count,
                )
                if any(legacy_request.errors):
                    num_bad += 1
                    self.migration_report.add_general_statistics(
                        i18n.t("Requests with valueErrors")
                    )
                    for error in legacy_request.errors:
                        self.migration_report.add("DiscardedRequests", f"{error[0]} - {error[1]}")
                        Helper.log_data_issue(
                            legacy_request.item_barcode,
                            f"{error[0]} - {error[1]}",
                            json.dumps(legacy_request.to_source_dict()),
                        )
                else:
                    self.migration_report.add_general_statistics(
                        i18n.t("Requests with valid source data")
                    )
                    yield legacy_request
            except ValueError as ve:
                logging.exception(ve)
        logging.info(
            f"Done validating {legacy_reques_count} "
            f"legacy requests with {num_bad} rotten apples"
        )
        if num_bad > 0 and (num_bad / legacy_reques_count) > 0.5:
            q = num_bad / legacy_reques_count
            logging.error("%s percent of requests failed to validate.", (q * 100))
            self.migration_report.log_me()
            logging.critical("Halting...")
            sys.exit(1)


def timings(t0, t0func, num_objects):
    avg = num_objects / (time.time() - t0)
    elapsed = time.time() - t0
    elapsed_func = time.time() - t0func
    return (
        f"Total objects: {num_objects}\tTotal elapsed: {elapsed:.2f}\t"
        f"Average per object: {avg:.2f}\tElapsed this time: {elapsed_func:.2f}"
    )



import json
import logging
import i18n


class Helper:
    @staticmethod
    def print_mapping_report(
        report_file, total_records: int, mapped_folio_fields, mapped_legacy_fields
    ):
        details_start = (
            "<details><summary>" + i18n.t("Click to expand field report") + "</summary>\n\n"
        )
        details_end = "</details>\n"
        report_file.write("\n## " + i18n.t("Mapped FOLIO fields") + "\n")
        # report_file.write(f"{blurbs[header]}\n")

        d_sorted = {k: mapped_folio_fields[k] for k in sorted(mapped_folio_fields)}
        report_file.write(details_start)
        columns = [i18n.t("FOLIO Field"), i18n.t("Mapped"), i18n.t("Unmapped")]
        report_file.write(" | ".join(columns) + "\n")
        report_file.write("|".join(len(columns) * ["---"]) + "\n")
        for k, v in d_sorted.items():
            unmapped = max(total_records - v[0], 0)
            mapped = v[0]
            mp = mapped / total_records if total_records else 0
            mapped_per = "{:.0%}".format(max(mp, 0))
            up = unmapped / total_records if total_records else 0
            unmapped_per = "{:.0%}".format(max(up, 0))
            report_file.write(
                f"{k} | {max(mapped, 0):,} ({mapped_per}) | {unmapped:,} ({unmapped_per}) \n"
            )
        report_file.write(details_end)

        report_file.write("\n## " + i18n.t("Mapped Legacy fields") + "\n")
        # report_file.write(f"{blurbs[header]}\n")

        d_sorted = {k: mapped_legacy_fields[k] for k in sorted(mapped_legacy_fields)}
        report_file.write(details_start)
        columns = [i18n.t("Legacy Field"), i18n.t("Present"), i18n.t("Mapped"), i18n.t("Unmapped")]
        report_file.write("|".join(columns) + "\n")
        report_file.write("|".join(len(columns) * ["---"]) + "\n")
        for k, v in d_sorted.items():
            present = v[0]
            present_per = "{:.1%}".format(present / total_records if total_records else 0)
            unmapped = present - v[1]
            mapped = v[1]
            mp = mapped / total_records if total_records else 0
            mapped_per = "{:.0%}".format(max(mp, 0))
            report_file.write(
                f"{k} | {max(present, 0):,} ({present_per}) | {max(mapped, 0):,} "
                f"({mapped_per}) | {unmapped:,}  \n"
            )
        report_file.write(details_end)

    @staticmethod
    def log_data_issue(index_or_id, message, legacy_value):
        logging.log(26, "DATA ISSUE\t%s\t%s\t%s", index_or_id, message, legacy_value)

    @staticmethod
    def write_to_file(file, folio_record):
        """Writes record to file.

        Args:
            file (_type_): _description_
            folio_record (_type_): _description_
        """
        file.write(f"{json.dumps(folio_record)}\n")



'''Main "script."'''
import csv
import logging
import i18n
from typing import Annotated
from typing import List

from folio_uuid.folio_namespaces import FOLIONamespaces
from pydantic import Field

from folio_migration_tools.custom_exceptions import TransformationProcessError
from folio_migration_tools.helper import Helper
from folio_migration_tools.library_configuration import FileDefinition
from folio_migration_tools.library_configuration import HridHandling
from folio_migration_tools.library_configuration import LibraryConfiguration
from folio_migration_tools.marc_rules_transformation.rules_mapper_holdings import (
    RulesMapperHoldings,
)
from folio_migration_tools.migration_tasks.migration_task_base import MigrationTaskBase
from folio_migration_tools.task_configuration import AbstractTaskConfiguration


class HoldingsMarcTransformer(MigrationTaskBase):
    class TaskConfiguration(AbstractTaskConfiguration):
        name: Annotated[
            str,
            Field(
                description=(
                    "Name of this migration task. The name is being used to call the specific "
                    "task, and to distinguish tasks of similar types"
                )
            ),
        ]
        migration_task_type: Annotated[
            str,
            Field(
                title="Migration task type",
                description=("The type of migration task you want to perform"),
            ),
        ]
        files: Annotated[
            List[FileDefinition],
            Field(
                title="Source files", description=("List of MARC21 files with holdings records")
            ),
        ]
        hrid_handling: Annotated[
            HridHandling,
            Field(
                title="HRID Handling",
                description=(
                    "Setting to default will make FOLIO generate HRIDs and move the existing "
                    "001:s into a 035, concatenated with the 003. Choosing preserve001 means "
                    "the 001:s will remain in place, and that they will also become the HRIDs"
                ),
            ),
        ] = HridHandling.default
        deactivate035_from001: Annotated[
            bool,
            Field(
                title="Create 035 from 001 and 003",
                description=(
                    "This deactivates the FOLIO default functionality of moving the previous 001 "
                    "into a 035, prefixed with the value from 003"
                ),
            ),
        ] = False
        holdings_type_uuid_for_boundwiths: Annotated[
            str,
            Field(
                title="Holdings Type for Boundwith Holdings",
                description=(
                    "UUID for a Holdings type (set in Settings->Inventory) "
                    "for Bound-with Holdings)"
                ),
            ),
        ] = ""
        boundwith_relationship_file_path: Annotated[
            str,
            Field(
                title="Boundwith relationship file path",
                description=(
                    "Path to a file outlining Boundwith relationships, in the style of Voyager."
                    " A TSV file with MFHD_ID and BIB_ID headers and values"
                ),
            ),
        ] = ""
        create_source_records: Annotated[
            bool,
            Field(
                title="Create source records",
                description=(
                    "Controls wheter or not to retain the MARC records in "
                    "Source Record Storage."
                ),
            ),
        ] = True
        update_hrid_settings: Annotated[
            bool,
            Field(
                title="Update HRID settings",
                description="At the end of the run, update FOLIO with the HRID settings",
            ),
        ] = True
        reset_hrid_settings: Annotated[
            bool,
            Field(
                title="Reset HRID settings",
                description=(
                    "Setting to true means the task will "
                    "reset the HRID counters for this particular record type"
                ),
            ),
        ] = False
        legacy_id_marc_path: Annotated[
            str,
            Field(
                title="Path to legacy id in the records",
                description=(
                    "The path to the field where the legacy id is located. "
                    "Example syntax: '001' or '951$c'"
                ),
            ),
        ]
        deduplicate_holdings_statements: Annotated[
            bool,
            Field(
                title="Deduplicate holdings statements",
                description=(
                    "If set to False, duplicate holding statements within the same record will "
                    "remain in place"
                ),
            ),
        ] = True
        location_map_file_name: Annotated[
            str,
            Field(
                title="Path to location map file",
                description="Must be a TSV file located in the mapping_files folder",
            ),
        ]
        default_call_number_type_name: Annotated[
            str,
            Field(
                title="Default callnumber type name",
                description="The name of the callnumber type that will be used as fallback",
            ),
        ]
        fallback_holdings_type_id: Annotated[
            str,
            Field(
                title="Fallback holdings type id",
                description="The UUID of the Holdings type that will be used for unmapped values",
            ),
        ]

    @staticmethod
    def get_object_type() -> FOLIONamespaces:
        return FOLIONamespaces.holdings

    def __init__(
        self,
        task_config: TaskConfiguration,
        library_config: LibraryConfiguration,
        use_logging: bool = True,
    ):
        csv.register_dialect("tsv", delimiter="\t")
        super().__init__(library_config, task_config, use_logging)
        self.task_config = task_config
        self.holdings_types = list(
            self.folio_client.folio_get_all("/holdings-types", "holdingsTypes")
        )
        self.default_holdings_type = next(
            (
                h
                for h in self.holdings_types
                if h["id"] == self.task_config.fallback_holdings_type_id
            ),
            {"name": ""},
        )
        if not self.default_holdings_type:
            raise TransformationProcessError(
                "",
                (
                    f"Holdings type with ID {self.task_config.fallback_holdings_type_id}"
                    " not found in FOLIO."
                ),
            )
        logging.info(
            "%s will be used as default holdings type",
            self.default_holdings_type.get("name", ""),
        )

        # Load Boundwith relationship map
        self.boundwith_relationship_map = []
        if self.task_config.boundwith_relationship_file_path:
            with open(
                self.folder_structure.legacy_records_folder
                / self.task_config.boundwith_relationship_file_path
            ) as boundwith_relationship_file:
                self.boundwith_relationship_map = list(
                    csv.DictReader(boundwith_relationship_file, dialect="tsv")
                )
            logging.info(
                "Rows in Bound with relationship map: %s", len(self.boundwith_relationship_map)
            )

        location_map_path = (
            self.folder_structure.mapping_files_folder / self.task_config.location_map_file_name
        )
        with open(location_map_path) as location_map_file:
            self.location_map = list(csv.DictReader(location_map_file, dialect="tsv"))
            logging.info("Locations in map: %s", len(self.location_map))

        self.check_source_files(
            self.folder_structure.legacy_records_folder, self.task_config.files
        )
        self.instance_id_map = self.load_id_map(self.folder_structure.instance_id_map_path, True)
        self.mapper = RulesMapperHoldings(
            self.folio_client,
            self.location_map,
            self.task_config,
            self.library_configuration,
            self.instance_id_map,
            self.boundwith_relationship_map,
        )
        if (
            self.task_configuration.reset_hrid_settings
            and self.task_configuration.update_hrid_settings
        ):
            self.mapper.hrid_handler.reset_holdings_hrid_counter()
        logging.info("%s Instance ids in map", len(self.instance_id_map))
        logging.info("Init done")

    def do_work(self):
        self.do_work_marc_transformer()

    def wrap_up(self):
        logging.info("Done. Transformer Wrapping up...")
        self.extradata_writer.flush()
        self.processor.wrap_up()
        with open(self.folder_structure.migration_reports_file, "w+") as report_file:
            self.mapper.migration_report.write_migration_report(
                i18n.t("Bibliographic records transformation report"),
                report_file,
                self.start_datetime,
            )
            Helper.print_mapping_report(
                report_file,
                self.mapper.parsed_records,
                self.mapper.mapped_folio_fields,
                self.mapper.mapped_legacy_fields,
            )

        logging.info(
            "Done. Transformation report written to %s",
            self.folder_structure.migration_reports_file.name,
        )

        self.clean_out_empty_logs()



import json
import logging
import sys
from datetime import datetime
from datetime import timezone
from typing import Set
from uuid import uuid4
import i18n

from folio_uuid.folio_uuid import FOLIONamespaces
from folioclient import FolioClient

from folio_migration_tools.custom_exceptions import TransformationProcessError
from folio_migration_tools.custom_exceptions import TransformationRecordFailedError
from folio_migration_tools.helper import Helper
from folio_migration_tools.library_configuration import FileDefinition
from folio_migration_tools.library_configuration import LibraryConfiguration
from folio_migration_tools.mapping_file_transformation.mapping_file_mapper_base import (
    MappingFileMapperBase,
)
from folio_migration_tools.mapping_file_transformation.ref_data_mapping import (
    RefDataMapping,
)


class ItemMapper(MappingFileMapperBase):
    def __init__(
        self,
        folio_client: FolioClient,
        items_map,
        material_type_map,
        loan_type_map,
        location_map,
        call_number_type_map,
        holdings_id_map,
        statistical_codes_map,
        item_statuses_map,
        temporary_loan_type_mapping,
        temporary_location_mapping,
        library_configuration: LibraryConfiguration,
        boundwith_relationship_map,
    ):
        item_schema = folio_client.get_item_schema()
        super().__init__(
            folio_client,
            item_schema,
            items_map,
            statistical_codes_map,
            FOLIONamespaces.items,
            library_configuration,
        )
        self.item_schema = self.folio_client.get_item_schema()
        self.items_map = items_map
        self.holdings_id_map = holdings_id_map
        self.unique_barcodes: Set[str] = set()
        self.status_mapping: dict = {}
        if temporary_loan_type_mapping:
            self.temp_loan_type_mapping = RefDataMapping(
                self.folio_client,
                "/loan-types",
                "loantypes",
                temporary_loan_type_mapping,
                "name",
                "TemporaryLoanTypeMapping",
            )
        self.temp_location_mapping = None
        if temporary_location_mapping:
            self.temp_location_mapping = RefDataMapping(
                self.folio_client,
                "/locations",
                "locations",
                temporary_location_mapping,
                "code",
                "TemporaryLocationMapping",
            )

        if item_statuses_map:
            self.setup_status_mapping(item_statuses_map)
        if call_number_type_map:
            self.call_number_mapping = RefDataMapping(
                self.folio_client,
                "/call-number-types",
                "callNumberTypes",
                call_number_type_map,
                "name",
                "CallNumberTypeMapping",
            )
        self.loan_type_mapping = RefDataMapping(
            self.folio_client,
            "/loan-types",
            "loantypes",
            loan_type_map,
            "name",
            "PermanentLoanTypeMapping",
        )
        self.boundwith_relationship_map = self.setup_boundwith_relationship_map(
            boundwith_relationship_map
        )

        self.material_type_mapping = RefDataMapping(
            self.folio_client,
            "/material-types",
            "mtypes",
            material_type_map,
            "name",
            "MaterialTypeMapping",
        )

        self.location_mapping = RefDataMapping(
            self.folio_client,
            "/locations",
            "locations",
            location_map,
            "code",
            "LocationMapping",
        )

    def perform_additional_mappings(self, folio_rec, file_def):
        self.handle_suppression(folio_rec, file_def)

    def handle_suppression(self, folio_record, file_def: FileDefinition):
        folio_record["discoverySuppress"] = file_def.discovery_suppressed
        self.migration_report.add(
            "Suppression",
            i18n.t("Suppressed from discovery") + f' = {folio_record["discoverySuppress"]}',
        )

    def setup_status_mapping(self, item_statuses_map):
        statuses = self.item_schema["properties"]["status"]["properties"]["name"]["enum"]
        for mapping in item_statuses_map:
            if "folio_name" not in mapping:
                logging.critical("folio_name is not a column in the status mapping file")
                sys.exit(1)
            elif "legacy_code" not in mapping:
                logging.critical("legacy_code is not a column in the status mapping file")
                sys.exit(1)
            elif mapping["folio_name"] not in statuses:
                logging.critical(
                    "%s in the mapping file is not a FOLIO item status",
                    mapping["folio_name"],
                )
                sys.exit(1)
            elif mapping["legacy_code"] == "*":
                logging.critical(
                    "* in status mapping not allowed. Available will be the default mapping. "
                    "Please remove the row with the *"
                )
                sys.exit(1)
            elif not all(mapping.values()):
                logging.critical("empty value in mapping %s. Check mapping file", mapping.values())
                sys.exit(1)
            else:
                self.status_mapping = {
                    v["legacy_code"]: v["folio_name"] for v in item_statuses_map
                }
        logging.info(json.dumps(statuses, indent=True))

    def get_prop(self, legacy_item, folio_prop_name, index_or_id, schema_default_value):
        if folio_prop_name == "permanentLocationId":
            return self.get_mapped_ref_data_value(
                self.location_mapping,
                legacy_item,
                folio_prop_name,
                index_or_id,
                False,
            )
        elif folio_prop_name == "temporaryLocationId":
            if not self.temp_location_mapping:
                raise TransformationProcessError(
                    "Temporary location is mapped, but there is no "
                    "temporary location mapping file referenced in configuration"
                )
            temp_loc = self.get_mapped_ref_data_value(
                self.temp_location_mapping,
                legacy_item,
                folio_prop_name,
                index_or_id,
                True,
            )
            self.migration_report.add("TemporaryLocationMapping", f"{temp_loc}")
            return temp_loc
        elif folio_prop_name == "materialTypeId":
            return self.get_mapped_ref_data_value(
                self.material_type_mapping,
                legacy_item,
                folio_prop_name,
                index_or_id,
            )
        elif folio_prop_name == "itemLevelCallNumberTypeId":
            return self.get_item_level_call_number_type_id(
                legacy_item, folio_prop_name, index_or_id
            )
        elif folio_prop_name == "status.date":
            return datetime.now(timezone.utc).isoformat()
        elif folio_prop_name == "temporaryLoanTypeId":
            ltid = self.get_mapped_ref_data_value(
                self.temp_loan_type_mapping,
                legacy_item,
                folio_prop_name,
                index_or_id,
                True,
            )
            self.migration_report.add("TemporaryLoanTypeMapping", f"{folio_prop_name} -> {ltid}")
            return ltid
        elif folio_prop_name == "permanentLoanTypeId":
            return self.get_mapped_ref_data_value(
                self.loan_type_mapping, legacy_item, folio_prop_name, index_or_id
            )
        elif folio_prop_name.startswith("statisticalCodeIds"):
            statistical_code_id = self.get_statistical_code(
                legacy_item, folio_prop_name, index_or_id
            )
            self.migration_report.add(
                "StatisticalCodeMapping",
                f"{folio_prop_name} -> {statistical_code_id}",
            )
            return statistical_code_id

        mapped_value = super().get_prop(
            legacy_item, folio_prop_name, index_or_id, schema_default_value
        )
        if folio_prop_name == "status.name":
            return self.transform_status(mapped_value)
        elif folio_prop_name == "barcode":
            barcode = mapped_value
            if barcode.strip() and barcode in self.unique_barcodes:
                Helper.log_data_issue(index_or_id, "Duplicate barcode", mapped_value)
                self.migration_report.add_general_statistics(i18n.t("Duplicate barcodes"))
                return f"{barcode}-{uuid4()}"
            else:
                if barcode.strip():
                    self.unique_barcodes.add(barcode)
                return barcode
        elif folio_prop_name == "holdingsRecordId":
            if mapped_value in self.holdings_id_map:
                return self.holdings_id_map[mapped_value][1]
            elif f"{self.bib_id_template}{mapped_value}" in self.holdings_id_map:
                return self.holdings_id_map[f"{self.bib_id_template}{mapped_value}"][1]
            self.migration_report.add_general_statistics(
                i18n.t("Records failed because of failed holdings"),
            )
            s = (
                "Holdings id referenced in legacy item "
                "was not found amongst transformed Holdings records"
            )
            raise TransformationRecordFailedError(index_or_id, s, mapped_value)
        elif mapped_value:
            return mapped_value
        else:
            self.migration_report.add("UnmappedProperties", f"{folio_prop_name}")
            return ""

    def get_item_level_call_number_type_id(self, legacy_item, folio_prop_name: str, index_or_id):
        if self.call_number_mapping:
            return self.get_mapped_ref_data_value(
                self.call_number_mapping, legacy_item, index_or_id, folio_prop_name
            )
        self.migration_report.add(
            "CallNumberTypeMapping",
            i18n.t("Mapping not setup"),
        )
        return ""

    def transform_status(self, legacy_value):
        status = self.status_mapping.get(legacy_value, "Available")
        self.migration_report.add("StatusMapping", f"'{legacy_value}' -> {status}")
        return status



'''Main "script."'''
import csv
import ctypes
import json
import logging
import sys
import time
import traceback
import uuid
import i18n
from typing import Annotated
from typing import List
from typing import Optional

from folio_uuid.folio_namespaces import FOLIONamespaces
from pydantic import Field

from folio_migration_tools.custom_exceptions import TransformationProcessError
from folio_migration_tools.custom_exceptions import TransformationRecordFailedError
from folio_migration_tools.helper import Helper
from folio_migration_tools.library_configuration import FileDefinition
from folio_migration_tools.library_configuration import HridHandling
from folio_migration_tools.library_configuration import LibraryConfiguration
from folio_migration_tools.mapping_file_transformation.item_mapper import ItemMapper
from folio_migration_tools.mapping_file_transformation.mapping_file_mapper_base import (
    MappingFileMapperBase,
)
from folio_migration_tools.marc_rules_transformation.hrid_handler import HRIDHandler
from folio_migration_tools.migration_tasks.migration_task_base import MigrationTaskBase
from folio_migration_tools.task_configuration import AbstractTaskConfiguration

csv.field_size_limit(int(ctypes.c_ulong(-1).value // 2))


class ItemsTransformer(MigrationTaskBase):
    class TaskConfiguration(AbstractTaskConfiguration):
        name: str
        migration_task_type: str
        hrid_handling: HridHandling
        files: List[FileDefinition]
        items_mapping_file_name: str
        location_map_file_name: str
        default_call_number_type_name: str
        temp_location_map_file_name: Optional[str] = ""
        material_types_map_file_name: str
        loan_types_map_file_name: str
        temp_loan_types_map_file_name: Optional[str] = ""
        statistical_codes_map_file_name: Optional[str] = ""
        item_statuses_map_file_name: str
        call_number_type_map_file_name: str
        reset_hrid_settings: Optional[bool] = False
        update_hrid_settings: Annotated[
            bool,
            Field(
                title="Update HRID settings",
                description="At the end of the run, update FOLIO with the HRID settings",
            ),
        ] = True
        boundwith_relationship_file_path: Annotated[
            str,
            Field(
                title="Boundwith relationship file path",
                description=(
                    "Path to a file outlining Boundwith relationships, in the style of Voyager."
                    " A TSV file with MFHD_ID and BIB_ID headers and values"
                ),
            ),
        ] = ""

    @staticmethod
    def get_object_type() -> FOLIONamespaces:
        return FOLIONamespaces.items

    def __init__(
        self,
        task_config: TaskConfiguration,
        library_config: LibraryConfiguration,
        use_logging: bool = True,
    ):
        csv.register_dialect("tsv", delimiter="\t")
        super().__init__(library_config, task_config, use_logging)
        self.task_config = task_config
        self.check_source_files(
            self.folder_structure.legacy_records_folder, self.task_config.files
        )
        self.total_records = 0
        self.folio_keys = []
        self.items_map = self.setup_records_map(
            self.folder_structure.mapping_files_folder / self.task_config.items_mapping_file_name
        )
        self.folio_keys = MappingFileMapperBase.get_mapped_folio_properties_from_map(
            self.items_map
        )
        if any(k for k in self.folio_keys if k.startswith("statisticalCodeIds")):
            statcode_mapping = self.load_ref_data_mapping_file(
                "statisticalCodeIds",
                self.folder_structure.mapping_files_folder
                / self.task_config.statistical_codes_map_file_name,
                self.folio_keys,
                False,
            )
        else:
            statcode_mapping = None

        if (
            self.folder_structure.mapping_files_folder
            / self.task_config.temp_loan_types_map_file_name
        ).is_file():
            temporary_loan_type_mapping = self.load_ref_data_mapping_file(
                "temporaryLoanTypeId",
                self.folder_structure.temp_loan_type_map_path,
                self.folio_keys,
            )
        else:
            logging.info(
                "%s not found. No temporary loan type mapping will be performed",
                self.folder_structure.temp_loan_type_map_path,
            )
            temporary_loan_type_mapping = None
        # Load Boundwith relationship map
        self.boundwith_relationship_map = []
        if self.task_config.boundwith_relationship_file_path:
            with open(
                self.folder_structure.data_folder
                / FOLIONamespaces.holdings.name
                / self.task_config.boundwith_relationship_file_path
            ) as boundwith_relationship_file:
                self.boundwith_relationship_map = list(
                    csv.DictReader(boundwith_relationship_file, dialect="tsv")
                )
            logging.info(
                "Rows in Bound with relationship map: %s", len(self.boundwith_relationship_map)
            )

        if (
            self.folder_structure.mapping_files_folder
            / self.task_config.temp_location_map_file_name
        ).is_file():
            temporary_location_mapping = self.load_ref_data_mapping_file(
                "temporaryLocationId",
                self.folder_structure.mapping_files_folder
                / self.task_config.temp_location_map_file_name,
                self.folio_keys,
            )
        else:
            logging.info(
                "%s not found. No temporary location mapping will be performed",
                self.task_config.temp_location_map_file_name,
            )
            temporary_location_mapping = None
        self.mapper = ItemMapper(
            self.folio_client,
            self.items_map,
            self.load_ref_data_mapping_file(
                "materialTypeId",
                self.folder_structure.mapping_files_folder
                / self.task_config.material_types_map_file_name,
                self.folio_keys,
            ),
            self.load_ref_data_mapping_file(
                "permanentLoanTypeId",
                self.folder_structure.mapping_files_folder
                / self.task_config.loan_types_map_file_name,
                self.folio_keys,
            ),
            self.load_ref_data_mapping_file(
                "permanentLocationId",
                self.folder_structure.mapping_files_folder
                / self.task_config.location_map_file_name,
                self.folio_keys,
            ),
            self.load_ref_data_mapping_file(
                "itemLevelCallNumberTypeId",
                self.folder_structure.mapping_files_folder
                / self.task_config.call_number_type_map_file_name,
                self.folio_keys,
                False,
            ),
            self.load_id_map(self.folder_structure.holdings_id_map_path),
            statcode_mapping,
            self.load_ref_data_mapping_file(
                "status.name",
                self.folder_structure.item_statuses_map_path,
                self.folio_keys,
                False,
            ),
            temporary_loan_type_mapping,
            temporary_location_mapping,
            self.library_configuration,
            self.boundwith_relationship_map,
        )
        if (
            self.task_configuration.reset_hrid_settings
            and self.task_configuration.update_hrid_settings
        ):
            hrid_handler = HRIDHandler(
                self.folio_client, HridHandling.default, self.mapper.migration_report, True
            )
            hrid_handler.reset_item_hrid_counter()

        logging.info("Init done")

    def do_work(self):
        logging.info("Starting....")
        with open(self.folder_structure.created_objects_path, "w+") as results_file:
            for file_def in self.task_config.files:
                try:
                    self.process_single_file(file_def, results_file)
                except Exception as exception:
                    error_str = f"\n\nProcessing of {file_def.file_name} failed:\n{exception}."
                    logging.exception(error_str, stack_info=True)
                    logging.fatal("Check source files for empty rows or missing reference data.")
                    self.mapper.migration_report.add(
                        "FailedFiles", f"{file_def.file_name} - {exception}"
                    )
                    logging.fatal(error_str)
                    sys.exit(1)
        logging.info(
            f"processed {self.total_records:,} records in {len(self.task_config.files)} files"
        )

    def process_single_file(self, file_def: FileDefinition, results_file):
        full_path = self.folder_structure.legacy_records_folder / file_def.file_name
        logging.info("Processing %s", full_path)
        records_in_file = 0
        with open(full_path, encoding="utf-8-sig") as records_file:
            self.mapper.migration_report.add_general_statistics(
                i18n.t("Number of files processed")
            )
            start = time.time()
            for idx, record in enumerate(self.mapper.get_objects(records_file, full_path)):
                try:
                    if idx == 0:
                        logging.info("First legacy record:")
                        logging.info(json.dumps(record, indent=4))
                        self.mapper.verify_legacy_record(record, idx)
                    folio_rec, legacy_id = self.mapper.do_map(
                        record, f"row {idx}", FOLIONamespaces.items
                    )

                    self.mapper.perform_additional_mappings(folio_rec, file_def)
                    self.handle_circiulation_notes(folio_rec, self.folio_client.current_user)
                    self.handle_notes(folio_rec)
                    if folio_rec["holdingsRecordId"] in self.mapper.boundwith_relationship_map:
                        for idx, instance_id in enumerate(
                            self.mapper.boundwith_relationship_map.get(
                                folio_rec["holdingsRecordId"]
                            )
                        ):
                            if idx == 0:
                                bw_id = folio_rec["holdingsRecordId"]
                            else:
                                bw_id = self.mapper.generate_boundwith_holding_uuid(
                                    folio_rec["holdingsRecordId"], instance_id
                                )
                            self.mapper.create_and_write_boundwith_part(legacy_id, bw_id)
                    if idx == 0:
                        logging.info("First FOLIO record:")
                        logging.info(json.dumps(folio_rec, indent=4))
                    # TODO: turn this into a asynchrounous task
                    Helper.write_to_file(results_file, folio_rec)
                    self.mapper.migration_report.add_general_statistics(
                        i18n.t("Number of records written to disk")
                    )
                    self.mapper.report_folio_mapping(folio_rec, self.mapper.schema)
                except TransformationProcessError as process_error:
                    self.mapper.handle_transformation_process_error(idx, process_error)
                except TransformationRecordFailedError as data_error:
                    self.mapper.handle_transformation_record_failed_error(idx, data_error)
                except AttributeError as attribute_error:
                    traceback.print_exc()
                    logging.fatal(attribute_error)
                    logging.info("Quitting...")
                    sys.exit(1)
                except Exception as excepion:
                    self.mapper.handle_generic_exception(idx, excepion)
                self.mapper.migration_report.add(
                    "GeneralStatistics",
                    i18n.t("Number of Legacy items in %{container}", container=file_def),
                )
                self.mapper.migration_report.add_general_statistics(
                    i18n.t("Number of Legacy items in total")
                )
                self.print_progress(idx, start)
                records_in_file = idx + 1

            logging.info(
                f"Done processing {file_def} containing {records_in_file:,} records. "
                f"Total records processed: {records_in_file:,}"
            )
        self.total_records += records_in_file

    @staticmethod
    def handle_notes(folio_object):
        if folio_object.get("notes", []):
            filtered_notes = []
            for note_obj in folio_object.get("notes", []):
                if not note_obj.get("itemNoteTypeId", ""):
                    raise TransformationProcessError(
                        folio_object.get("legacyIds", ""),
                        "Missing note type id mapping",
                        json.dumps(note_obj),
                    )
                elif note_obj.get("note", "") and note_obj.get("itemNoteTypeId", ""):
                    filtered_notes.append(note_obj)
            if filtered_notes:
                folio_object["notes"] = filtered_notes
            else:
                del folio_object["notes"]

    @staticmethod
    def handle_circiulation_notes(folio_rec, current_user_uuid):
        if not folio_rec.get("circulationNotes", []):
            return
        filtered_notes = []
        for circ_note in folio_rec.get("circulationNotes", []):
            if circ_note.get("noteType", "") not in ["Check in", "Check out"]:
                raise TransformationProcessError(
                    "", "Circulation Note types are not mapped correclty"
                )
            if circ_note.get("note", ""):
                circ_note["id"] = str(uuid.uuid4())
                circ_note["source"] = {
                    "id": current_user_uuid,
                    "personal": {"lastName": "Data", "firstName": "Migration"},
                }
                filtered_notes.append(circ_note)
        if filtered_notes:
            folio_rec["circulationNotes"] = filtered_notes
        else:
            del folio_rec["circulationNotes"]

    def wrap_up(self):
        logging.info("Done. Transformer wrapping up...")
        self.extradata_writer.flush()
        with open(self.folder_structure.migration_reports_file, "w") as migration_report_file:
            self.mapper.migration_report.write_migration_report(
                i18n.t("Item transformation report"),
                migration_report_file,
                self.mapper.start_datetime,
            )
            Helper.print_mapping_report(
                migration_report_file,
                self.total_records,
                self.mapper.mapped_folio_fields,
                self.mapper.mapped_legacy_fields,
            )
        self.clean_out_empty_logs()
        logging.info("All done!")



import copy
import json
import logging
import re
import time
import i18n
from typing import Set

import httpx
from folioclient import FolioClient
from httpx import HTTPError

from folio_migration_tools.helper import Helper
from folio_migration_tools.migration_report import MigrationReport
from folio_migration_tools.transaction_migration.legacy_loan import LegacyLoan
from folio_migration_tools.transaction_migration.legacy_request import LegacyRequest
from folio_migration_tools.transaction_migration.transaction_result import (
    TransactionResult,
)

date_time_format = "%Y-%m-%dT%H:%M:%S.%f+0000"


class CirculationHelper:
    def __init__(
        self,
        folio_client: FolioClient,
        service_point_id,
        migration_report: MigrationReport,
    ):
        self.folio_client = folio_client
        self.service_point_id = service_point_id
        self.missing_patron_barcodes: Set[str] = set()
        self.missing_item_barcodes: Set[str] = set()
        self.migration_report: MigrationReport = migration_report

    def get_user_by_barcode(self, user_barcode):
        if user_barcode in self.missing_patron_barcodes:
            self.migration_report.add_general_statistics(
                i18n.t("Users already detected as missing")
            )
            logging.info("User is already detected as missing")
            return {}
        user_path = f"/users?query=barcode=={user_barcode}"
        try:
            users = self.folio_client.folio_get(user_path, "users")
            if any(users):
                return users[0]
            self.missing_patron_barcodes.add(user_barcode)
            return {}
        except Exception as ee:
            logging.error(f"{ee} {user_path}")
            return {}

    def get_item_by_barcode(self, item_barcode):
        if item_barcode in self.missing_item_barcodes:
            self.migration_report.add_general_statistics(
                i18n.t("Items already detected as missing")
            )
            logging.info("Item is already detected as missing")
            return {}
        item_path = f"/item-storage/items?query=barcode=={item_barcode}"
        try:
            item = self.folio_client.folio_get(item_path, "items")
            if any(item):
                return item[0]
            self.missing_item_barcodes.add(item_barcode)
            return {}
        except Exception as ee:
            logging.error(f"{ee} {item_path}")
            return {}

    def is_checked_out(self, legacy_loan: LegacyLoan) -> bool:
        """Makes a deeper check to find out if the loan is already processed.
        Looks up the item id, and then searches Loan Storage for any open loans.
        If there are open loans, returns True. Else False.

        Args:
            legacy_loan (LegacyLoan): _description_

        Returns:
            bool: _description_
        """
        if item := self.get_item_by_barcode(legacy_loan.item_barcode):
            if self.get_active_loan_by_item_id(item["id"]):
                return True
        return False

    def get_active_loan_by_item_id(self, item_id: str) -> dict:
        """Queries FOLIO for the first open loan.

        Args:
            item_id (str): the item ID. A uuid string.

        Returns:
            dict: The open loan, if found. Else an empty dictionary
        """
        loan_path = f'/loan-storage/loans?query=(itemId=="{item_id}")'
        try:
            loans = self.folio_client.folio_get(loan_path, "loans")
            return next((loan for loan in loans if loan["status"]["name"] == "Open"), {})
        except Exception as ee:
            logging.error(f"{ee} {loan_path}")
            return {}

    def get_holding_by_uuid(self, holdings_uuid):
        holdings_path = f"/holdings-storage/holdings/{holdings_uuid}"
        try:
            return self.folio_client.folio_get_single_object(holdings_path)
        except Exception as ee:
            logging.error(f"{ee} {holdings_path}")
            return {}

    def check_out_by_barcode(self, legacy_loan: LegacyLoan) -> TransactionResult:
        """Checks out a legacy loan using the Endpoint /circulation/check-out-by-barcode
        Adds all possible overrides in order to make the transaction go through

        Args:
            legacy_loan (LegacyLoan): _description_

        Returns:
            TransactionResult: _description_
        """

        t0_function = time.time()
        data = {
            "itemBarcode": legacy_loan.item_barcode,
            "userBarcode": legacy_loan.patron_barcode,
            "loanDate": legacy_loan.out_date.isoformat(),
            "servicePointId": legacy_loan.service_point_id,
            "overrideBlocks": {
                "itemNotLoanableBlock": {"dueDate": legacy_loan.due_date.isoformat()},
                "patronBlock": {},
                "itemLimitBlock": {},
                "comment": "Migrated from legacy system",
            },
        }
        path = "/circulation/check-out-by-barcode"
        url = f"{self.folio_client.okapi_url}{path}"
        try:
            if legacy_loan.patron_barcode in self.missing_patron_barcodes:
                error_message = i18n.t("Patron barcode already detected as missing")
                logging.error(
                    f"{error_message} Patron barcode: {legacy_loan.patron_barcode} "
                    f"Item Barcode:{legacy_loan.item_barcode}"
                )
                return TransactionResult(False, False, "", error_message, error_message)
            req = httpx.post(url, headers=self.folio_client.okapi_headers, json=data, timeout=None)
            if req.status_code == 422:
                error_message_from_folio = json.loads(req.text)["errors"][0]["message"]
                stat_message = error_message_from_folio
                error_message = error_message_from_folio
                if "has the item status" in error_message_from_folio:
                    stat_message = re.findall(
                        r"(?<=has the item status\s).*(?=\sand cannot be checked out)",
                        error_message_from_folio,
                    )[0]
                    error_message = (
                        f"{stat_message} for item with barcode {legacy_loan.item_barcode}"
                    )
                    return TransactionResult(
                        False,
                        True,
                        None,
                        error_message_from_folio,
                        stat_message,
                    )
                elif "No item with barcode" in error_message_from_folio:
                    error_message = f"No item with barcode {legacy_loan.item_barcode} in FOLIO"
                    stat_message = "Item barcode not in FOLIO"
                    self.missing_item_barcodes.add(legacy_loan.item_barcode)
                    return TransactionResult(
                        False,
                        False,
                        None,
                        error_message_from_folio,
                        stat_message,
                    )

                elif "find user with matching barcode" in error_message_from_folio:
                    self.missing_patron_barcodes.add(legacy_loan.patron_barcode)
                    error_message = f"No patron with barcode {legacy_loan.patron_barcode} in FOLIO"
                    stat_message = i18n.t("Patron barcode not in FOLIO")
                    return TransactionResult(
                        False,
                        False,
                        None,
                        error_message_from_folio,
                        stat_message,
                    )
                elif "Cannot check out item that already has an open" in error_message_from_folio:
                    return TransactionResult(
                        False,
                        False,
                        None,
                        error_message_from_folio,
                        error_message_from_folio,
                    )
                logging.error(
                    f"{error_message} "
                    f"Patron barcode: {legacy_loan.patron_barcode} "
                    f"Item Barcode:{legacy_loan.item_barcode}"
                )
                self.migration_report.add("Details", stat_message)
                return TransactionResult(
                    False, True, None, error_message, f"Check out error: {stat_message}"
                )
            elif req.status_code == 201:
                stats = "Successfully checked out by barcode"
                logging.debug(
                    "%s (item barcode %s}) in %ss",
                    stats,
                    legacy_loan.item_barcode,
                    f"{(time.time() - t0_function):.2f}",
                )
                return TransactionResult(True, False, json.loads(req.text), "", stats)
            elif req.status_code == 204:
                stats = "Successfully checked out by barcode"
                logging.debug(
                    "%s (item barcode %s) %s",
                    stats,
                    legacy_loan.item_barcode,
                    req.status_code,
                )
                return TransactionResult(True, False, None, "", stats)
            else:
                req.raise_for_status()
        except HTTPError:
            logging.exception(
                "%s\tPOST FAILED %s\n\t%s\n\t%s",
                req.status_code,
                url,
                json.dumps(data),
                req.text,
            )
            return TransactionResult(
                False,
                False,
                None,
                "5XX",
                i18n.t("Failed checkout http status %{code}", code=req.status_code),
            )

    @staticmethod
    def create_request(
        folio_client: FolioClient, legacy_request: LegacyRequest, migration_report: MigrationReport
    ):
        try:
            path = "/circulation/httpx"
            url = f"{folio_client.okapi_url}{path}"
            data = legacy_request.serialize()
            data["requestProcessingParameters"] = {
                "overrideBlocks": {
                    "itemNotLoanableBlock": {
                        "dueDate": legacy_request.request_expiration_date.isoformat()
                    },
                    "patronBlock": {},
                    "itemLimitBlock": {},
                    "comment": "Migrated from legacy system",
                }
            }
            req = httpx.post(url, headers=folio_client.okapi_headers, json=data, timeout=None)
            logging.debug(f"POST {req.status_code}\t{url}\t{json.dumps(data)}")
            if str(req.status_code) == "422":
                message = json.loads(req.text)["errors"][0]["message"]
                logging.error(f"{message}")
                migration_report.add_general_statistics(message)
                return False
            else:
                req.raise_for_status()
                logging.info(
                    "%s Successfully created %s",
                    req.status_code,
                    legacy_request.request_type,
                )
                return True
        except Exception as exception:
            logging.error(exception, exc_info=True)
            migration_report.add("Details", exception)
            Helper.log_data_issue(
                legacy_request.item_barcode,
                exception,
                json.dumps(legacy_request.to_source_dict()),
            )
            return False

    def load_migrated_user_barcodes(self, user_barcodes, patron_files, folder_structure):
        if any(patron_files):
            for filedef in patron_files:
                my_path = folder_structure.results_folder / filedef.file_name
                with open(my_path) as patron_file:
                    for row in patron_file:
                        rec = json.loads(row)
                        user_barcodes.add(rec.get("barcode", "None"))
            logging.info("Loaded %s barcodes from users", len(user_barcodes))

    def load_migrated_item_barcodes(self, item_barcodes, item_files, folder_structure):
        if any(item_files):
            for filedef in item_files:
                my_path = folder_structure.results_folder / filedef.file_name
                with open(my_path) as item_file:
                    for row in item_file:
                        rec = json.loads(row)
                        item_barcodes.add(rec.get("barcode", "None"))
            logging.info("Loaded %s barcodes from items", len(item_barcodes))

    @staticmethod
    def extend_open_loan(folio_client: FolioClient, loan, extension_due_date, extend_out_date):
        try:
            loan_to_put = copy.deepcopy(loan)
            del loan_to_put["metadata"]
            loan_to_put["dueDate"] = extension_due_date.isoformat()
            loan_to_put["loanDate"] = extend_out_date.isoformat()
            url = f"{folio_client.okapi_url}/circulation/loans/{loan_to_put['id']}"

            req = httpx.put(
                url, headers=folio_client.okapi_headers, json=loan_to_put, timeout=None
            )
            logging.info(
                "%s\tPUT Extend loan %s to %s\t %s",
                req.status_code,
                loan_to_put["id"],
                loan_to_put["dueDate"],
                url,
            )
            if str(req.status_code) == "422":
                logging.error(
                    "%s\t%s",
                    json.loads(req.text)["errors"][0]["message"],
                    json.dumps(loan_to_put),
                )
                return False
            else:
                req.raise_for_status()
                logging.info("%s Successfully Extended loan", req.status_code)
            return True
        except Exception:
            logging.exception(
                "PUT FAILED Extend loan to %s\t %s\t%s",
                loan_to_put["dueDate"],
                url,
                json.dumps(loan_to_put),
            )
            return False



import copy
import json
import logging
import sys
import time
import traceback
import i18n
from datetime import datetime
from typing import Annotated
from typing import List
from uuid import uuid4

import httpx
from folio_uuid.folio_namespaces import FOLIONamespaces
from pydantic import Field

from folio_migration_tools.custom_exceptions import TransformationProcessError
from folio_migration_tools.custom_exceptions import TransformationRecordFailedError
from folio_migration_tools.library_configuration import FileDefinition
from folio_migration_tools.library_configuration import LibraryConfiguration
from folio_migration_tools.migration_report import MigrationReport
from folio_migration_tools.migration_tasks.migration_task_base import MigrationTaskBase
from folio_migration_tools.task_configuration import AbstractTaskConfiguration


def write_failed_batch_to_file(batch, file):
    for record in batch:
        file.write(f"{json.dumps(record)}\n")


class BatchPoster(MigrationTaskBase):
    """Batchposter

    Args:
        MigrationTaskBase (_type_): _description_

    Raises:
        ee: _description_
        TransformationRecordFailedError: _description_
        TransformationProcessError: _description_
        TransformationRecordFailedError: _description_
        TransformationRecordFailedError: _description_
        TransformationProcessError: _description_

    Returns:
        _type_: _description_
    """

    class TaskConfiguration(AbstractTaskConfiguration):
        name: str
        migration_task_type: str
        object_type: str
        files: List[FileDefinition]
        batch_size: int
        rerun_failed_records: Annotated[
            bool,
            Field(
                description=(
                    "Toggles whether or not BatchPoster should try to rerun failed batches or "
                    "just leave the failing records on disk."
                )
            ),
        ] = True
        use_safe_inventory_endpoints: Annotated[
            bool,
            Field(
                description=(
                    "Toggles the use of the safe/unsafe Inventory storage endpoints. "
                    "Unsafe circumvents the Optimistic locking in FOLIO. Defaults to "
                    "True (using the 'safe' options)"
                )
            ),
        ] = True

    @staticmethod
    def get_object_type() -> FOLIONamespaces:
        return FOLIONamespaces.other

    def __init__(
        self,
        task_config: TaskConfiguration,
        library_config: LibraryConfiguration,
        use_logging: bool = True,
    ):
        super().__init__(library_config, task_config, use_logging)
        self.migration_report = MigrationReport()
        self.performing_rerun = False
        self.failed_ids: list = []
        self.first_batch = True
        self.api_info = get_api_info(
            self.task_configuration.object_type,
            self.task_configuration.use_safe_inventory_endpoints,
        )
        self.snapshot_id = str(uuid4())
        self.failed_objects: list = []
        self.batch_size = self.task_configuration.batch_size
        logging.info("Batch size is %s", self.batch_size)
        self.processed = 0
        self.failed_batches = 0
        self.users_created = 0
        self.users_updated = 0
        self.users_per_group: dict = {}
        self.failed_fields: set = set()
        self.num_failures = 0
        self.num_posted = 0
        self.okapi_headers = self.folio_client.okapi_headers
        self.http_client = None

    def do_work(self):
        with httpx.Client(timeout=None, headers=self.folio_client.okapi_headers) as httpx_client:
            self.http_client = httpx_client
            try:
                batch = []
                if self.task_configuration.object_type == "SRS":
                    self.create_snapshot()
                with open(self.folder_structure.failed_recs_path, "w") as failed_recs_file:
                    for file_def in self.task_configuration.files:
                        path = self.folder_structure.results_folder / file_def.file_name
                        with open(path) as rows:
                            logging.info("Running %s", path)
                            last_row = ""
                            for self.processed, row in enumerate(rows, start=1):
                                last_row = row
                                if row.strip():
                                    try:
                                        if self.task_configuration.object_type == "Extradata":
                                            self.post_extra_data(
                                                row, self.processed, failed_recs_file
                                            )
                                        elif not self.api_info["is_batch"]:
                                            self.post_single_records(
                                                row, self.processed, failed_recs_file
                                            )
                                        else:
                                            batch = self.post_record_batch(
                                                batch, failed_recs_file, row
                                            )
                                    except UnicodeDecodeError as unicode_error:
                                        self.handle_unicode_error(unicode_error, last_row)
                                    except TransformationProcessError as tpe:
                                        self.handle_generic_exception(
                                            tpe,
                                            last_row,
                                            batch,
                                            self.processed,
                                            failed_recs_file,
                                        )
                                        logging.critical("Halting %s", tpe)
                                        print(f"\n\t{tpe.message}")
                                        sys.exit(1)
                                    except TransformationRecordFailedError as exception:
                                        self.handle_generic_exception(
                                            exception,
                                            last_row,
                                            batch,
                                            self.processed,
                                            failed_recs_file,
                                        )
                                        batch = []

                    if self.task_configuration.object_type != "Extradata" and any(batch):
                        try:
                            self.post_batch(batch, failed_recs_file, self.processed)
                        except Exception as exception:
                            self.handle_generic_exception(
                                exception, last_row, batch, self.processed, failed_recs_file
                            )
                    logging.info("Done posting %s records. ", (self.processed))
            except Exception as ee:
                if self.task_configuration.object_type == "SRS":
                    self.commit_snapshot()
                raise ee

    def post_record_batch(self, batch, failed_recs_file, row):
        json_rec = json.loads(row.split("\t")[-1])
        if (
            self.task_configuration.object_type in ["Instances", "Holdings", "Items"]
            and not self.task_configuration.use_safe_inventory_endpoints
        ):
            self.migration_report.add_general_statistics(
                i18n.t("Set _version to -1 to enable upsert")
            )
            json_rec["_version"] = -1
        if self.task_configuration.object_type == "SRS":
            json_rec["snapshotId"] = self.snapshot_id
        if self.processed == 1:
            logging.info(json.dumps(json_rec, indent=True))
        batch.append(json_rec)
        if len(batch) == int(self.batch_size):
            self.post_batch(batch, failed_recs_file, self.processed)
            batch = []
        return batch

    def post_extra_data(self, row: str, num_records: int, failed_recs_file):
        (object_name, data) = row.split("\t")
        endpoint = get_extradata_endpoint(object_name, data)
        url = f"{self.folio_client.okapi_url}/{endpoint}"
        body = data
        response = self.post_objects(url, body)
        if response.status_code == 201:
            self.num_posted += 1
        elif response.status_code == 422:
            self.num_failures += 1
            error_msg = json.loads(response.text)["errors"][0]["message"]
            logging.error("Row %s\tHTTP %s\t %s", num_records, response.status_code, error_msg)
            if "id value already exists" not in json.loads(response.text)["errors"][0]["message"]:
                failed_recs_file.write(row)
        else:
            self.num_failures += 1
            logging.error("Row %s\tHTTP %s\t%s", num_records, response.status_code, response.text)
            failed_recs_file.write(row)
        if num_records % 50 == 0:
            logging.info(
                "%s records posted successfully. %s failed",
                self.num_posted,
                self.num_failures,
            )

    def post_single_records(self, row: str, num_records: int, failed_recs_file):
        if self.api_info["is_batch"]:
            raise TypeError("This record type supports batch processing, use post_batch method")
        api_endpoint = self.api_info.get("api_endpoint")
        url = f"{self.folio_client.okapi_url}{api_endpoint}"
        response = self.post_objects(url, row)
        if response.status_code == 201:
            self.num_posted += 1
        elif response.status_code == 422:
            self.num_failures += 1
            error_msg = json.loads(response.text)["errors"][0]["message"]
            logging.error("Row %s\tHTTP %s\t %s", num_records, response.status_code, error_msg)
            if "id value already exists" not in json.loads(response.text)["errors"][0]["message"]:
                failed_recs_file.write(row)
        else:
            self.num_failures += 1
            logging.error("Row %s\tHTTP %s\t%s", num_records, response.status_code, response.text)
            failed_recs_file.write(row)
        if num_records % 50 == 0:
            logging.info(
                "%s records posted successfully. %s failed",
                self.num_posted,
                self.num_failures,
            )

    def post_objects(self, url, body):
        if self.http_client and not self.http_client.is_closed:
            return self.http_client.post(url, data=body.encode("utf-8"))
        else:
            return httpx.post(
                url, headers=self.okapi_headers, data=body.encode("utf-8"), timeout=None
            )

    def handle_generic_exception(self, exception, last_row, batch, num_records, failed_recs_file):
        logging.error("%s", exception)
        self.migration_report.add("Details", i18n.t("Generic exceptions (see log for details)"))
        # logging.error("Failed row: %s", last_row)
        self.failed_batches += 1
        self.num_failures += len(batch)
        write_failed_batch_to_file(batch, failed_recs_file)
        logging.info("Resetting batch...Number of failed batches: %s", self.failed_batches)
        batch = []
        if self.failed_batches > 50000:
            logging.error("Exceeded number of failed batches at row %s", num_records)
            logging.critical("Halting")
            sys.exit(1)

    def handle_unicode_error(self, unicode_error, last_row):
        self.migration_report.add("Details", i18n.t("Encoding errors"))
        logging.info("=========ERROR==============")
        logging.info(
            "%s Posting failed. Encoding error reading file",
            unicode_error,
        )
        logging.info(
            "Failing row, either the one shown here or the next row in %s",
            self.task_configuration.file.file_name,
        )
        logging.info(last_row)
        logging.info("=========Stack trace==============")
        traceback.logging.info_exc()
        logging.info("=======================", flush=True)

    def post_batch(self, batch, failed_recs_file, num_records, recursion_depth=0):
        response = self.do_post(batch)
        if response.status_code == 201:
            logging.info(
                (
                    "Posting successful! Total rows: %s Total failed: %s "
                    "in %ss "
                    "Batch Size: %s Request size: %s "
                ),
                num_records,
                self.num_failures,
                response.elapsed.total_seconds(),
                len(batch),
                get_req_size(response),
            )
        elif response.status_code == 200:
            json_report = json.loads(response.text)
            self.users_created += json_report.get("createdRecords", 0)
            self.users_updated += json_report.get("updatedRecords", 0)
            self.num_posted = self.users_updated + self.users_created
            self.num_failures += json_report.get("failedRecords", 0)
            if json_report.get("failedRecords", 0) > 0:
                logging.error(
                    "%s users in batch failed to load",
                    json_report.get("failedRecords", 0),
                )
                write_failed_batch_to_file(batch, failed_recs_file)
            if json_report.get("failedUsers", []):
                logging.error("Errormessage: %s", json_report.get("error", []))
                for failed_user in json_report.get("failedUsers"):
                    logging.error(
                        "User failed. %s\t%s\t%s",
                        failed_user.get("username", ""),
                        failed_user.get("externalSystemId", ""),
                        failed_user.get("errorMessage", ""),
                    )
                    self.migration_report.add("Details", failed_user.get("errorMessage", ""))
            logging.info(
                (
                    "Posting successful! Total rows: %s Total failed: %s "
                    "created: %s updated: %s in %ss Batch Size: %s Request size: %s "
                    "Message from server: %s"
                ),
                num_records,
                self.num_failures,
                self.users_created,
                self.users_updated,
                response.elapsed.total_seconds(),
                len(batch),
                get_req_size(response),
                json_report.get("message", ""),
            )
        elif response.status_code == 422:
            resp = json.loads(response.text)
            raise TransformationRecordFailedError(
                "",
                f"HTTP {response.status_code}\t"
                f"Request size: {get_req_size(response)}"
                f"{datetime.utcnow().isoformat()} UTC\n",
                json.dumps(resp, indent=4),
            )
        elif response.status_code == 400:
            # Likely a json parsing error
            logging.error(response.text)
            raise TransformationProcessError("", "HTTP 400. Somehting is wrong. Quitting")
        elif self.task_configuration.object_type == "SRS" and response.status_code >= 500:
            logging.info(
                "Post failed. Size: %s Waiting 30s until reposting. Number of tries: %s of 5",
                get_req_size(response),
                recursion_depth,
            )
            logging.info(response.text)
            time.sleep(30)
            if recursion_depth > 4:
                raise TransformationRecordFailedError(
                    "",
                    f"HTTP {response.status_code}\t"
                    f"Request size: {get_req_size(response)}"
                    f"{datetime.utcnow().isoformat()} UTC\n",
                    response.text,
                )
            else:
                self.post_batch(batch, failed_recs_file, num_records, recursion_depth + 1)
        elif (
            response.status_code == 413 and "DB_ALLOW_SUPPRESS_OPTIMISTIC_LOCKING" in response.text
        ):
            logging.error(response.text)
            raise TransformationProcessError("", response.text, "")

        else:
            try:
                logging.info(response.text)
                resp = json.dumps(response, indent=4)
            except TypeError:
                resp = response
            except Exception:
                logging.exception("something unexpected happened")
                resp = response
            raise TransformationRecordFailedError(
                "",
                f"HTTP {response.status_code}\t"
                f"Request size: {get_req_size(response)}"
                f"{datetime.utcnow().isoformat()} UTC\n",
                resp,
            )

    def do_post(self, batch):
        path = self.api_info["api_endpoint"]
        url = self.folio_client.okapi_url + path
        if self.api_info["object_name"] == "users":
            payload = {self.api_info["object_name"]: list(batch), "totalRecords": len(batch)}
        elif self.api_info["total_records"]:
            payload = {"records": list(batch), "totalRecords": len(batch)}
        else:
            payload = {self.api_info["object_name"]: batch}
        if self.http_client and not self.http_client.is_closed:
            return self.http_client.post(url, json=payload)
        else:
            return httpx.post(url, headers=self.okapi_headers, json=payload, timeout=None)

    def wrap_up(self):
        logging.info("Done. Wrapping up")
        self.extradata_writer.flush()
        if self.task_configuration.object_type == "SRS":
            self.commit_snapshot()
        if self.task_configuration.object_type != "Extradata":
            logging.info(
                (
                    "Failed records: %s failed records in %s "
                    "failed batches. Failed records saved to %s"
                ),
                self.num_failures,
                self.failed_batches,
                self.folder_structure.failed_recs_path,
            )
        else:
            logging.info("Done posting %s records. %s failed", self.num_posted, self.num_failures)

        run = "second time" if self.performing_rerun else "first time"
        self.migration_report.set("GeneralStatistics", f"Records processed {run}", self.processed)
        self.migration_report.set("GeneralStatistics", f"Records posted {run}", self.num_posted)
        self.migration_report.set("GeneralStatistics", f"Failed to post {run}", self.num_failures)
        self.rerun_run()
        with open(self.folder_structure.migration_reports_file, "w+") as report_file:
            self.migration_report.write_migration_report(
                f"{self.task_configuration.object_type} loading report",
                report_file,
                self.start_datetime,
            )
        self.clean_out_empty_logs()

    def rerun_run(self):
        if self.task_configuration.rerun_failed_records and (self.num_failures > 0):
            logging.info(
                "Rerunning the %s failed records from the load with a batchsize of 1",
                self.num_failures,
            )
            try:
                self.task_configuration.batch_size = 1
                self.task_configuration.files = [
                    FileDefinition(file_name=str(self.folder_structure.failed_recs_path.name))
                ]
                temp_report = copy.deepcopy(self.migration_report)
                temp_start = self.start_datetime
                self.task_configuration.rerun_failed_records = False
                self.__init__(self.task_configuration, self.library_configuration)
                self.performing_rerun = True
                self.migration_report = temp_report
                self.start_datetime = temp_start
                self.do_work()
                self.wrap_up()
                logging.info("Done rerunning the posting")
            except Exception as ee:
                logging.exception("Occurred during rerun")
                raise TransformationProcessError("Error during rerun") from ee
        elif not self.task_configuration.rerun_failed_records and (self.num_failures > 0):
            logging.info(
                (
                    "Task configured to not rerun failed records. "
                    " File with failed records is located at %s"
                ),
                str(self.folder_structure.failed_recs_path),
            )

    def create_snapshot(self):
        snapshot = {
            "jobExecutionId": self.snapshot_id,
            "status": "PARSING_IN_PROGRESS",
            "processingStartedDate": datetime.utcnow().isoformat(timespec="milliseconds"),
        }
        try:
            url = f"{self.folio_client.okapi_url}/source-storage/snapshots"
            if self.http_client and not self.http_client.is_closed:
                res = self.http_client.post(url, json=snapshot)
            else:
                res = httpx.post(url, headers=self.okapi_headers, json=snapshot, timeout=None)
            res.raise_for_status()
            logging.info("Posted Snapshot to FOLIO: %s", json.dumps(snapshot, indent=4))
            get_url = f"{self.folio_client.okapi_url}/source-storage/snapshots/{self.snapshot_id}"
            getted = False
            while not getted:
                logging.info("Sleeping while waiting for the snapshot to get created")
                time.sleep(5)
                if self.http_client and not self.http_client.is_closed:
                    res = self.http_client.get(get_url)
                else:
                    res = httpx.get(get_url, headers=self.okapi_headers, timeout=None)
                if res.status_code == 200:
                    getted = True
                else:
                    logging.info(res.status_code)
        except Exception:
            logging.exception("Could not post the snapshot")
            sys.exit(1)

    def commit_snapshot(self):
        snapshot = {"jobExecutionId": self.snapshot_id, "status": "COMMITTED"}
        try:
            url = f"{self.folio_client.okapi_url}/source-storage/snapshots/{self.snapshot_id}"
            if self.http_client and not self.http_client.is_closed:
                res = self.http_client.put(url, json=snapshot)
            else:
                res = httpx.put(url, headers=self.okapi_headers, json=snapshot, timeout=None)
            res.raise_for_status()
            logging.info("Posted Committed snapshot to FOLIO: %s", json.dumps(snapshot, indent=4))
        except Exception:
            logging.exception(
                "Could not commit snapshot with id %s. Post this to /source-storage/snapshots/%s:",
                self.snapshot_id,
                self.snapshot_id,
            )
            logging.info("%s", json.dumps(snapshot, indent=4))
            sys.exit(1)


def get_api_info(object_type: str, use_safe: bool = True):
    choices = {
        "Extradata": {
            "object_name": "",
            "api_endpoint": "",
            "total_records": False,
            "addSnapshotId": False,
        },
        "Items": {
            "object_name": "items",
            "api_endpoint": (
                "/item-storage/batch/synchronous"
                if use_safe
                else "/item-storage/batch/synchronous-unsafe"
            ),
            "is_batch": True,
            "total_records": False,
            "addSnapshotId": False,
        },
        "Holdings": {
            "object_name": "holdingsRecords",
            "api_endpoint": (
                "/holdings-storage/batch/synchronous"
                if use_safe
                else "/holdings-storage/batch/synchronous-unsafe"
            ),
            "is_batch": True,
            "total_records": False,
            "addSnapshotId": False,
        },
        "Instances": {
            "object_name": "instances",
            "api_endpoint": (
                "/instance-storage/batch/synchronous"
                if use_safe
                else "/instance-storage/batch/synchronous-unsafe"
            ),
            "is_batch": True,
            "total_records": False,
            "addSnapshotId": False,
        },
        "Authorities": {
            "object_name": "",
            "api_endpoint": "/authority-storage/authorities",
            "is_batch": False,
            "total_records": False,
            "addSnapshotId": False,
        },
        "SRS": {
            "object_name": "records",
            "api_endpoint": "/source-storage/batch/records",
            "is_batch": True,
            "total_records": True,
            "addSnapshotId": True,
        },
        "Users": {
            "object_name": "users",
            "api_endpoint": "/user-import",
            "is_batch": True,
            "total_records": True,
            "addSnapshotId": False,
        },
        "Organizations": {
            "object_name": "",
            "api_endpoint": "/organizations/organizations",
            "is_batch": False,
            "total_records": False,
            "addSnapshotId": False,
        },
        "Orders": {
            "object_name": "",
            "api_endpoint": "/orders/composite-orders",
            "is_batch": False,
            "total_records": False,
            "addSnapshotId": False,
        },
    }

    try:
        return choices[object_type]
    except KeyError:
        key_string = ",".join(choices.keys())
        logging.error(f"Wrong type. Only one of {key_string} are allowed")
        logging.error("Halting")
        sys.exit(1)


def chunks(records, number_of_chunks):
    """Yield successive n-sized chunks from lst.

    Args:
        records (_type_): _description_
        number_of_chunks (_type_): _description_

    Yields:
        _type_: _description_
    """
    for i in range(0, len(records), number_of_chunks):
        yield records[i : i + number_of_chunks]


def get_extradata_endpoint(object_name: str, string_object: str):
    object_types = {
        "precedingSucceedingTitles": "preceding-succeeding-titles",
        "precedingTitles": "preceding-succeeding-titles",
        "succeedingTitles": "preceding-succeeding-titles",
        "boundwithPart": "inventory-storage/bound-with-parts",
        "notes": "notes",
        "course": "coursereserves/courses",
        "courselisting": "coursereserves/courselistings",
        "contacts": "organizations-storage/contacts",
        "interfaces": "organizations-storage/interfaces",
        "account": "accounts",
        "feefineaction": "feefineactions",
    }
    if object_name == "instructor":
        instructor = json.loads(string_object)
        return f'coursereserves/courselistings/{instructor["courseListingId"]}/instructors'

    if object_name == "interfaceCredential":
        credential = json.loads(string_object)
        return f'organizations-storage/interfaces/{credential["interfaceId"]}/credentials'

    return object_types[object_name]


def get_human_readable(size, precision=2):
    suffixes = ["B", "KB", "MB", "GB", "TB"]
    suffix_index = 0
    while size > 1024 and suffix_index < 4:
        suffix_index += 1  # increment the index of the suffix
        size = size / 1024.0  # apply the division
    return "%.*f%s" % (precision, size, suffixes[suffix_index])


def get_req_size(response: httpx.Response):
    size = response.request.method
    size += str(response.request.url)
    size += "\r\n".join(f"{k}{v}" for k, v in response.request.headers.items())
    size += response.request.content.decode("utf-8") or ""
    return get_human_readable(len(size.encode("utf-8")))
